#
# Autogenerated by Thrift Compiler (0.14.1)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:slots
#
import sys

from thrift.protocol.TProtocol import TProtocolException
from thrift.Thrift import TApplicationException
from thrift.Thrift import TException
from thrift.Thrift import TFrozenDict
from thrift.Thrift import TMessageType
from thrift.Thrift import TType
from thrift.transport import TTransport
from thrift.TRecursive import fix_spec

all_structs = []


class IsHealthyProbe(object):
    """
    The different types of probes supported by is_healthy endpoint.

    Please refer to Kubernetes' documentation for the differences between them:
    https://kubernetes.io/docs/tasks/configure-pod-container/configure-liveness-readiness-startup-probes/

    Your service should use Readiness probe as the fallback for unsupported probes.

    Note that the HTTP health check could use the string names of the probes,
    so changing the names, even without changing the numeric values,
    is considered as breaking change and should be avoided.


    """

    READINESS = 1
    LIVENESS = 2
    STARTUP = 3

    _VALUES_TO_NAMES = {
        1: "READINESS",
        2: "LIVENESS",
        3: "STARTUP",
    }

    _NAMES_TO_VALUES = {
        "READINESS": 1,
        "LIVENESS": 2,
        "STARTUP": 3,
    }


class ErrorCode(object):
    """
    The integer values within this enum correspond to HTTP status codes.

    HTTP layers can easily map errors to an appropriate status code.

    """

    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    PAYMENT_REQUIRED = 402
    FORBIDDEN = 403
    NOT_FOUND = 404
    CONFLICT = 409
    GONE = 410
    PRECONDITION_FAILED = 412
    PAYLOAD_TOO_LARGE = 413
    IM_A_TEAPOT = 418
    MISDIRECTED_REQUEST = 421
    UNPROCESSABLE_ENTITY = 422
    LOCKED = 423
    FAILED_DEPENDENCY = 424
    TOO_EARLY = 425
    PRECONDITION_REQUIRED = 428
    TOO_MANY_REQUESTS = 429
    REQUEST_HEADER_FIELDS_TOO_LARGE = 431
    UNAVAILABLE_FOR_LEGAL_REASONS = 451
    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501
    BAD_GATEWAY = 502
    SERVICE_UNAVAILABLE = 503
    TIMEOUT = 504
    INSUFFICIENT_STORAGE = 507
    LOOP_DETECTED = 508
    USER_DEFINED = 1000

    _VALUES_TO_NAMES = {
        400: "BAD_REQUEST",
        401: "UNAUTHORIZED",
        402: "PAYMENT_REQUIRED",
        403: "FORBIDDEN",
        404: "NOT_FOUND",
        409: "CONFLICT",
        410: "GONE",
        412: "PRECONDITION_FAILED",
        413: "PAYLOAD_TOO_LARGE",
        418: "IM_A_TEAPOT",
        421: "MISDIRECTED_REQUEST",
        422: "UNPROCESSABLE_ENTITY",
        423: "LOCKED",
        424: "FAILED_DEPENDENCY",
        425: "TOO_EARLY",
        428: "PRECONDITION_REQUIRED",
        429: "TOO_MANY_REQUESTS",
        431: "REQUEST_HEADER_FIELDS_TOO_LARGE",
        451: "UNAVAILABLE_FOR_LEGAL_REASONS",
        500: "INTERNAL_SERVER_ERROR",
        501: "NOT_IMPLEMENTED",
        502: "BAD_GATEWAY",
        503: "SERVICE_UNAVAILABLE",
        504: "TIMEOUT",
        507: "INSUFFICIENT_STORAGE",
        508: "LOOP_DETECTED",
        1000: "USER_DEFINED",
    }

    _NAMES_TO_VALUES = {
        "BAD_REQUEST": 400,
        "UNAUTHORIZED": 401,
        "PAYMENT_REQUIRED": 402,
        "FORBIDDEN": 403,
        "NOT_FOUND": 404,
        "CONFLICT": 409,
        "GONE": 410,
        "PRECONDITION_FAILED": 412,
        "PAYLOAD_TOO_LARGE": 413,
        "IM_A_TEAPOT": 418,
        "MISDIRECTED_REQUEST": 421,
        "UNPROCESSABLE_ENTITY": 422,
        "LOCKED": 423,
        "FAILED_DEPENDENCY": 424,
        "TOO_EARLY": 425,
        "PRECONDITION_REQUIRED": 428,
        "TOO_MANY_REQUESTS": 429,
        "REQUEST_HEADER_FIELDS_TOO_LARGE": 431,
        "UNAVAILABLE_FOR_LEGAL_REASONS": 451,
        "INTERNAL_SERVER_ERROR": 500,
        "NOT_IMPLEMENTED": 501,
        "BAD_GATEWAY": 502,
        "SERVICE_UNAVAILABLE": 503,
        "TIMEOUT": 504,
        "INSUFFICIENT_STORAGE": 507,
        "LOOP_DETECTED": 508,
        "USER_DEFINED": 1000,
    }


class IsHealthyRequest(object):
    """
    The arg struct for is_healthy endpoint.


    Attributes:
     - probe

    """

    __slots__ = ("probe",)

    def __init__(
        self,
        probe=None,
    ):
        self.probe = probe

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.probe = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin("IsHealthyRequest")
        if self.probe is not None:
            oprot.writeFieldBegin("probe", TType.I32, 1)
            oprot.writeI32(self.probe)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, getattr(self, key)) for key in self.__slots__]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for attr in self.__slots__:
            my_val = getattr(self, attr)
            other_val = getattr(other, attr)
            if my_val != other_val:
                return False
        return True

    def __ne__(self, other):
        return not (self == other)


class Error(TException):
    """
    Attributes:
     - code: A code describing the general nature of the error.
    This should be specified for all errors. This field uses
    the i32 type instead of the ErrorCode type in order to give
    developers an escape hatch to define their own error codes.
    Developers should do their best to avoid defining a custom
    error code. Developers should use a value higher than 1000
    when defining custom codes.
     - message: A human-readable error message. It should both explain the error
    and offer an actionable resolution to it, if applicable. It should
    be safe to desplay this message in a user-facing client.
     - details: A map of additional error information. This is most useful
    when there is a validation error. The server may use this map
    to return multiple errors. This should be safe for clients to
    display. Example:
        {
            "post.title": "This field is too long.",
            "post.kind": "This field is required."
        }
     - retryable: Server could choose to set this field to true to explicitly indicate
    that client shall retry this request, and false to explicitly indicate that
    client shall not retry this request. Unset means that it's up to the client
    to decide (using other information, for example the code) whether to retry
    this request.

    """

    __slots__ = (
        "code",
        "message",
        "details",
        "retryable",
    )

    def __init__(
        self,
        code=None,
        message=None,
        details=None,
        retryable=None,
    ):
        super(Error, self).__setattr__("code", code)
        super(Error, self).__setattr__("message", message)
        super(Error, self).__setattr__("details", details)
        super(Error, self).__setattr__("retryable", retryable)

    def __setattr__(self, *args):
        raise TypeError("can't modify immutable instance")

    def __delattr__(self, *args):
        raise TypeError("can't modify immutable instance")

    def __hash__(self):
        return hash(self.__class__) ^ hash(
            (
                self.code,
                self.message,
                self.details,
                self.retryable,
            )
        )

    @classmethod
    def read(cls, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and cls.thrift_spec is not None
        ):
            return iprot._fast_decode(None, iprot, [cls, cls.thrift_spec])
        iprot.readStructBegin()
        code = None
        message = None
        details = None
        retryable = None
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    code = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    message = (
                        iprot.readString().decode("utf-8", errors="replace")
                        if sys.version_info[0] == 2
                        else iprot.readString()
                    )
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.MAP:
                    details = {}
                    (_ktype1, _vtype2, _size0) = iprot.readMapBegin()
                    for _i4 in range(_size0):
                        _key5 = (
                            iprot.readString().decode("utf-8", errors="replace")
                            if sys.version_info[0] == 2
                            else iprot.readString()
                        )
                        _val6 = (
                            iprot.readString().decode("utf-8", errors="replace")
                            if sys.version_info[0] == 2
                            else iprot.readString()
                        )
                        details[_key5] = _val6
                    iprot.readMapEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.BOOL:
                    retryable = iprot.readBool()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()
        return cls(
            code=code,
            message=message,
            details=details,
            retryable=retryable,
        )

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin("Error")
        if self.code is not None:
            oprot.writeFieldBegin("code", TType.I32, 1)
            oprot.writeI32(self.code)
            oprot.writeFieldEnd()
        if self.message is not None:
            oprot.writeFieldBegin("message", TType.STRING, 2)
            oprot.writeString(
                self.message.encode("utf-8") if sys.version_info[0] == 2 else self.message
            )
            oprot.writeFieldEnd()
        if self.details is not None:
            oprot.writeFieldBegin("details", TType.MAP, 3)
            oprot.writeMapBegin(TType.STRING, TType.STRING, len(self.details))
            for kiter7, viter8 in self.details.items():
                oprot.writeString(kiter7.encode("utf-8") if sys.version_info[0] == 2 else kiter7)
                oprot.writeString(viter8.encode("utf-8") if sys.version_info[0] == 2 else viter8)
            oprot.writeMapEnd()
            oprot.writeFieldEnd()
        if self.retryable is not None:
            oprot.writeFieldBegin("retryable", TType.BOOL, 4)
            oprot.writeBool(self.retryable)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __str__(self):
        return repr(self)

    def __repr__(self):
        L = ["%s=%r" % (key, getattr(self, key)) for key in self.__slots__]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for attr in self.__slots__:
            my_val = getattr(self, attr)
            other_val = getattr(other, attr)
            if my_val != other_val:
                return False
        return True

    def __ne__(self, other):
        return not (self == other)


all_structs.append(IsHealthyRequest)
IsHealthyRequest.thrift_spec = (
    None,  # 0
    (
        1,
        TType.I32,
        "probe",
        None,
        None,
    ),  # 1
)
all_structs.append(Error)
Error.thrift_spec = (
    None,  # 0
    (
        1,
        TType.I32,
        "code",
        None,
        None,
    ),  # 1
    (
        2,
        TType.STRING,
        "message",
        "UTF8",
        None,
    ),  # 2
    (
        3,
        TType.MAP,
        "details",
        (TType.STRING, "UTF8", TType.STRING, "UTF8", False),
        None,
    ),  # 3
    (
        4,
        TType.BOOL,
        "retryable",
        None,
        None,
    ),  # 4
)
fix_spec(all_structs)
del all_structs



import re

from time import perf_counter
from typing import Any
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram
from sqlalchemy import create_engine
from sqlalchemy import event
from sqlalchemy.engine import Connection
from sqlalchemy.engine import Engine
from sqlalchemy.engine import ExceptionContext
from sqlalchemy.engine.interfaces import ExecutionContext
from sqlalchemy.engine.url import make_url
from sqlalchemy.orm import Session
from sqlalchemy.pool import QueuePool

from baseplate import _ExcInfo
from baseplate import Span
from baseplate import SpanObserver
from baseplate.clients import ContextFactory
from baseplate.lib import config
from baseplate.lib import metrics
from baseplate.lib.prometheus_metrics import default_latency_buckets
from baseplate.lib.secrets import SecretsStore


def engine_from_config(
    app_config: config.RawConfig,
    secrets: Optional[SecretsStore] = None,
    prefix: str = "database.",
    **kwargs: Any,
) -> Engine:
    """Make an :py:class:`~sqlalchemy.engine.Engine` from a configuration dictionary.

    The keys useful to :py:func:`engine_from_config` should be prefixed, e.g.
    ``database.url``, etc. The ``prefix`` argument specifies the prefix used to
    filter keys.

    Supported keys:

    * ``url``: the connection URL to the database, passed to
        :py:func:`~sqlalchemy.engine.url.make_url` to create the
        :py:class:`~sqlalchemy.engine.url.URL` used to connect to the database.
    * ``credentials_secret`` (optional): the key used to retrieve the database
        credentials from ``secrets`` as a :py:class:`~baseplate.lib.secrets.CredentialSecret`.
        If this is supplied, any credentials given in ``url`` we be replaced by
        these.
    * ``pool_recycle`` (optional): this setting causes the pool to recycle connections after
        the given number of seconds has passed. It defaults to -1, or no timeout.
    * ``pool_pre_ping`` (optional): when set to true, this setting causes
        sqlalchemy to perform a liveness-check query each time a connection is
        checked out of the pool.  If the liveness-check fails, the connection
        is gracefully recycled.  This ensures severed connections are handled
        more gracefully, at the cost of doing a `SELECT 1` at the start of each
        checkout. When used, this obviates most of the reasons you might use
        pool_recycle, and as such they shouldn't normally be used
        simultaneously.  Requires SQLAlchemy 1.3.
    * ``pool_size`` (optional) : The number of connections that can be saved in the pool.
    * ``max_overflow`` (optional) : Max connections that can be opened beyond the pool size.

    """
    assert prefix.endswith(".")
    parser = config.SpecParser(
        {
            "url": config.String,
            "credentials_secret": config.Optional(config.String),
            "pool_recycle": config.Optional(config.Integer),
            "pool_pre_ping": config.Optional(config.Boolean),
            "pool_size": config.Optional(config.Integer),
            "max_overflow": config.Optional(config.Integer),
        }
    )
    options = parser.parse(prefix[:-1], app_config)
    url = make_url(options.url)

    if options.pool_recycle is not None:
        kwargs.setdefault("pool_recycle", options.pool_recycle)

    if options.pool_pre_ping is not None:
        kwargs.setdefault("pool_pre_ping", options.pool_pre_ping)

    if options.pool_size is not None:
        kwargs.setdefault("pool_size", options.pool_size)

    if options.max_overflow is not None:
        kwargs.setdefault("max_overflow", options.max_overflow)

    if options.credentials_secret:
        if not secrets:
            raise TypeError("'secrets' is required if 'credentials_secret' is set")
        credentials = secrets.get_credentials(options.credentials_secret)

        # support sqlalchemy 1.4+ where URL is immutable
        # https://docs.sqlalchemy.org/en/14/changelog/migration_14.html#the-url-object-is-now-immutable
        if hasattr(url, "set"):
            url = url.set(username=credentials.username, password=credentials.password)
        else:
            url.username = credentials.username
            url.password = credentials.password

    return create_engine(url, **kwargs)


class SQLAlchemySession(config.Parser):
    """Configure a SQLAlchemy Session.

    This is meant to be used with
    :py:meth:`baseplate.Baseplate.configure_context`.

    See :py:func:`engine_from_config` for available configuration settings.

    :param secrets: Required if configured to use credentials to talk to the database.

    """

    def __init__(self, secrets: Optional[SecretsStore] = None, **kwargs: Any):
        self.secrets = secrets
        self.kwargs = kwargs

    def parse(
        self, key_path: str, raw_config: config.RawConfig
    ) -> "SQLAlchemySessionContextFactory":
        engine = engine_from_config(
            raw_config, secrets=self.secrets, prefix=f"{key_path}.", **self.kwargs
        )
        return SQLAlchemySessionContextFactory(engine, key_path)


Parameters = Optional[Union[Dict[str, Any], Sequence[Any]]]


SAFE_TRACE_ID = re.compile("^[A-Za-z0-9_-]+$")


class SQLAlchemyEngineContextFactory(ContextFactory):
    """SQLAlchemy core engine context factory.

    This factory will attach a SQLAlchemy :py:class:`sqlalchemy.engine.Engine`
    to an attribute on the :py:class:`~baseplate.RequestContext`. All cursor
    (query) execution will automatically record diagnostic information.

    Additionally, the trace and span ID will be added as a comment to the text
    of the SQL statement. This is to aid correlation of queries with requests.

    .. seealso::

        The engine is the low-level SQLAlchemy API. If you want to use the ORM,
        consider using
        :py:class:`~baseplate.clients.sqlalchemy.SQLAlchemySessionContextFactory`
        instead.

    :param engine: A configured SQLAlchemy engine.

    """

    PROM_PREFIX = "sql_client"
    PROM_POOL_PREFIX = f"{PROM_PREFIX}_pool"
    PROM_POOL_LABELS = ["sql_pool"]

    max_connections_gauge = Gauge(
        f"{PROM_POOL_PREFIX}_max_size",
        "Maximum number of connections allowed in this pool",
        PROM_POOL_LABELS,
    )

    checked_out_connections_gauge = Gauge(
        f"{PROM_POOL_PREFIX}_active_connections",
        "Number of connections in use by this pool (checked out + overflow)",
        PROM_POOL_LABELS,
    )
    checked_in_connections_gauge = Gauge(
        f"{PROM_POOL_PREFIX}_idle_connections",
        "Number of connections not in use by this pool (unused pool connections)",
        PROM_POOL_LABELS,
    )

    PROM_LABELS = [
        "sql_pool",
        "sql_address",
        "sql_database",
    ]

    latency_seconds = Histogram(
        f"{PROM_PREFIX}_latency_seconds",
        "Latency histogram of calls to database",
        PROM_LABELS + ["sql_success"],
        buckets=default_latency_buckets,
    )

    active_requests = Gauge(
        f"{PROM_PREFIX}_active_requests",
        "total requests that are in-flight",
        PROM_LABELS,
    )

    requests_total = Counter(
        f"{PROM_PREFIX}_requests_total",
        "Total number of sql requests",
        PROM_LABELS + ["sql_success"],
    )

    def __init__(self, engine: Engine, name: str = "sqlalchemy"):
        self.engine = engine.execution_options()
        self.name = name
        event.listen(self.engine, "before_cursor_execute", self.on_before_execute, retval=True)
        event.listen(self.engine, "after_cursor_execute", self.on_after_execute)
        event.listen(self.engine, "handle_error", self.on_error)
        self.time_started = 0.0

    def report_runtime_metrics(self, batch: metrics.Client) -> None:
        pool = self.engine.pool
        if not isinstance(pool, QueuePool):
            return

        self.max_connections_gauge.labels(self.name).set(pool.size())
        self.checked_out_connections_gauge.labels(self.name).set(pool.checkedout())
        self.checked_in_connections_gauge.labels(self.name).set(pool.checkedin())

        batch.gauge("pool.size").replace(pool.size())
        batch.gauge("pool.open_and_available").replace(pool.checkedin())
        batch.gauge("pool.in_use").replace(pool.checkedout())
        batch.gauge("pool.overflow").replace(max(pool.overflow(), 0))

    def make_object_for_context(self, name: str, span: Span) -> Engine:
        engine = self.engine.execution_options(context_name=name, server_span=span)
        return engine

    # pylint: disable=unused-argument, too-many-arguments
    def on_before_execute(
        self,
        conn: Connection,
        cursor: Any,
        statement: str,
        parameters: Parameters,
        context: Optional[ExecutionContext],
        executemany: bool,
    ) -> Tuple[str, Parameters]:
        """Handle the engine's before_cursor_execute event."""
        labels = {
            "sql_pool": self.name,
            "sql_address": conn.engine.url.host,
            "sql_database": conn.engine.url.database,
        }
        self.active_requests.labels(**labels).inc()
        self.time_started = perf_counter()

        context_name = conn._execution_options["context_name"]
        server_span = conn._execution_options["server_span"]

        trace_name = f"{context_name}.execute"
        span = server_span.make_child(trace_name)
        span.set_tag("statement", statement[:1021] + "..." if len(statement) > 1024 else statement)
        span.start()

        conn.info["span"] = span

        # add a comment to the sql statement with the trace and span ids
        # this is useful for slow query logs and active query views
        if SAFE_TRACE_ID.match(span.trace_id) and SAFE_TRACE_ID.match(span.id):
            annotated_statement = f"{statement} -- trace:{span.trace_id},span:{span.id}"
        else:
            annotated_statement = f"{statement} -- invalid trace id"

        return annotated_statement, parameters

    # pylint: disable=unused-argument, too-many-arguments
    def on_after_execute(
        self,
        conn: Connection,
        cursor: Any,
        statement: str,
        parameters: Parameters,
        context: Optional[ExecutionContext],
        executemany: bool,
    ) -> None:
        """Handle the event which happens after successful cursor execution."""
        conn.info["span"].finish()
        conn.info["span"] = None

        labels = {
            "sql_pool": self.name,
            "sql_address": conn.engine.url.host,
            "sql_database": conn.engine.url.database,
        }

        self.active_requests.labels(**labels).dec()
        self.requests_total.labels(**labels, sql_success="true").inc()
        self.latency_seconds.labels(**labels, sql_success="true").observe(
            perf_counter() - self.time_started
        )

    def on_error(self, context: ExceptionContext) -> None:
        """Handle the event which happens on exceptions during execution."""
        exc_info = (type(context.original_exception), context.original_exception, None)
        context.connection.info["span"].finish(exc_info=exc_info)
        context.connection.info["span"] = None

        labels = {
            "sql_pool": self.name,
            "sql_address": context.connection.engine.url.host,
            "sql_database": context.connection.engine.url.database,
        }

        self.active_requests.labels(**labels).dec()
        self.requests_total.labels(**labels, sql_success="false").inc()
        self.latency_seconds.labels(**labels, sql_success="false").observe(
            perf_counter() - self.time_started
        )


class SQLAlchemySessionContextFactory(SQLAlchemyEngineContextFactory):
    """SQLAlchemy ORM session context factory.

    This factory will attach a new SQLAlchemy
    :py:class:`sqlalchemy.orm.session.Session` to an attribute on the
    :py:class:`~baseplate.RequestContext`. All cursor (query) execution will
    automatically record diagnostic information.

    The session will be automatically closed, but not committed or rolled back,
    at the end of each request.

    .. seealso::

        The session is part of the high-level SQLAlchemy ORM API. If you want
        to do raw queries, consider using
        :py:class:`~baseplate.clients.sqlalchemy.SQLAlchemyEngineContextFactory`
        instead.

    :param engine: A configured SQLAlchemy engine.

    """

    def make_object_for_context(self, name: str, span: Span) -> Session:
        engine = super().make_object_for_context(name, span)
        session = Session(bind=engine)
        span.register(SQLAlchemySessionSpanObserver(session))
        return session


class SQLAlchemySessionSpanObserver(SpanObserver):
    """Automatically close the session at the end of each request."""

    def __init__(self, session: Session):
        self.session = session

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        self.session.close()



import importlib
import logging

import gevent
import pytest
import requests

from pyramid.config import Configurator
from pyramid.httpexceptions import HTTPNoContent

from baseplate import Baseplate
from baseplate.clients.requests import ExternalRequestsClient
from baseplate.clients.requests import InternalRequestsClient
from baseplate.frameworks.pyramid import BaseplateConfigurator
from baseplate.frameworks.pyramid import StaticTrustHandler
from baseplate.lib import config
from baseplate.server import make_listener
from baseplate.server.wsgi import make_server

from . import TestBaseplateObserver

logger = logging.getLogger(__name__)


@pytest.fixture
def gevent_socket():
    try:
        gevent.monkey.patch_socket()
        yield
    finally:
        import socket

        importlib.reload(socket)
        gevent.monkey.saved.clear()


@pytest.fixture
def http_server(gevent_socket):
    class HttpServer:
        def __init__(self, address):
            self.url = f"http://{address[0]}:{address[1]}/"
            self.requests = []

        def handle_request(self, request):
            self.requests.append(request)
            return HTTPNoContent()

    server_bind_endpoint = config.Endpoint("127.0.0.1:0")
    listener = make_listener(server_bind_endpoint)
    server_address = listener.getsockname()
    http_server = HttpServer(server_address)

    baseplate = Baseplate()
    trust_handler = StaticTrustHandler(trust_headers=True)
    baseplate_configurator = BaseplateConfigurator(baseplate, header_trust_handler=trust_handler)
    configurator = Configurator()
    configurator.include(baseplate_configurator.includeme)
    configurator.add_route("test_view", "/")
    configurator.add_view(http_server.handle_request, route_name="test_view", renderer="json")
    wsgi_app = configurator.make_wsgi_app()

    server = make_server({"stop_timeout": "1 millisecond"}, listener, wsgi_app)
    server_greenlet = gevent.spawn(server.serve_forever)
    try:
        yield http_server
    finally:
        server_greenlet.kill()


@pytest.mark.parametrize("client_cls", [InternalRequestsClient, ExternalRequestsClient])
@pytest.mark.parametrize("client_name", [None, "", "complex.client$name"])
@pytest.mark.parametrize("method", ["DELETE", "GET", "HEAD", "OPTIONS", "PATCH", "PUT", "POST"])
def test_client_makes_client_span(client_cls, client_name, method, http_server):
    baseplate = Baseplate(
        {"myclient.filter.ip_allowlist": "127.0.0.0/8", "myclient.filter.port_denylist": "0"}
    )
    if client_name is None:
        baseplate.configure_context({"myclient": client_cls()})
    else:
        baseplate.configure_context({"myclient": client_cls(client_name=client_name)})

    observer = TestBaseplateObserver()
    baseplate.register(observer)

    with baseplate.server_context("test") as context:
        fn = getattr(context.myclient, method.lower())
        response = fn(http_server.url)

    assert response.status_code == 204

    server_span_observer = observer.children[0]
    assert len(server_span_observer.children) == 1

    client_span_observer = server_span_observer.children[0]
    assert client_span_observer.span.name == "myclient.request"
    assert client_span_observer.on_start_called
    assert client_span_observer.on_finish_called
    assert client_span_observer.on_finish_exc_info is None
    assert client_span_observer.tags["http.url"] == http_server.url
    assert client_span_observer.tags["http.method"] == method.lower()
    assert client_span_observer.tags["http.status_code"] == 204
    assert client_span_observer.tags["http.slug"] == (
        client_name if client_name is not None else "myclient"
    )


@pytest.mark.parametrize("client_cls", [InternalRequestsClient, ExternalRequestsClient])
def test_connection_error(client_cls):
    baseplate = Baseplate(
        {"myclient.filter.ip_allowlist": "127.0.0.0/8", "myclient.filter.port_denylist": "0"}
    )
    baseplate.configure_context({"myclient": client_cls()})

    observer = TestBaseplateObserver()
    baseplate.register(observer)

    bogus_url = "http://localhost:1/"
    with pytest.raises(requests.exceptions.ConnectionError):
        with baseplate.server_context("test") as context:
            context.myclient.get(bogus_url)

    server_span_observer = observer.children[0]
    assert len(server_span_observer.children) == 1

    client_span_observer = server_span_observer.children[0]
    assert client_span_observer.span.name == "myclient.request"
    assert client_span_observer.on_start_called
    assert client_span_observer.on_finish_called
    assert client_span_observer.on_finish_exc_info is not None
    assert client_span_observer.tags["http.url"] == bogus_url
    assert client_span_observer.tags["http.method"] == "GET".lower()
    assert "http.status_code" not in client_span_observer.tags


def test_internal_client_sends_headers(http_server):
    baseplate = Baseplate()
    baseplate.configure_context({"internal": InternalRequestsClient()})

    with baseplate.server_context("test") as context:
        setattr(context, "raw_edge_context", b"test payload")

        response = context.internal.get(http_server.url)

        assert response.status_code == 204
        assert response.text == ""
        assert http_server.requests[0].method == "GET"
        assert http_server.requests[0].span.trace_id == context.span.trace_id
        assert http_server.requests[0].span.parent_id == context.span.id
        assert http_server.requests[0].span.id != context.span.id
        assert http_server.requests[0].raw_edge_context == b"test payload"


def test_external_client_doesnt_send_headers(http_server):
    baseplate = Baseplate(
        {"external.filter.ip_allowlist": "127.0.0.0/8", "external.filter.port_denylist": "0"}
    )
    baseplate.configure_context({"external": ExternalRequestsClient()})

    with baseplate.server_context("test") as context:
        setattr(context, "raw_edge_context", b"test payload")

        response = context.external.get(http_server.url)

        assert response.status_code == 204
        assert response.text == ""
        assert http_server.requests[0].method == "GET"
        assert "X-Trace" not in http_server.requests[0].headers
        assert "X-Parent" not in http_server.requests[0].headers
        assert "X-Span" not in http_server.requests[0].headers
        assert "X-Edge-Request" not in http_server.requests[0].headers



"""Write a file's contents to a node in ZooKeeper."""
import argparse
import configparser
import difflib
import logging
import sys

from typing import BinaryIO

from kazoo.client import KazooClient
from kazoo.exceptions import BadVersionError
from kazoo.exceptions import NoNodeError

from baseplate.lib.live_data.zookeeper import zookeeper_client_from_config
from baseplate.lib.secrets import secrets_store_from_config
from baseplate.server import EnvironmentInterpolation


logger = logging.getLogger(__name__)


class WriterError(Exception):
    pass


class NodeDoesNotExistError(WriterError):
    def __init__(self) -> None:
        super().__init__(
            "Target node does not exist. Please create it with appropriate ACLs first."
        )


class UnexpectedChangeError(WriterError):
    def __init__(self) -> None:
        super().__init__("The data in ZooKeeper changed unexpectedly.")


def write_file_to_zookeeper(zookeeper: KazooClient, source_file: BinaryIO, dest_path: str) -> bool:
    logger.info("Writing to %s in ZooKeeper...", dest_path)

    try:
        current_data, stat = zookeeper.get(dest_path)
        current_version = stat.version
    except NoNodeError:
        raise NodeDoesNotExistError

    new_data = source_file.read()

    if current_data == new_data:
        logger.info("No changes detected. Not writing.")
        return False

    try:
        current_text = current_data.decode("utf8")
        new_text = new_data.decode("utf8")
    except UnicodeDecodeError:
        logger.info("Skipping diff, data appears to be binary.")
    else:
        diff = difflib.unified_diff(current_text.splitlines(), new_text.splitlines())

        for line in diff:
            logger.info(line)

    try:
        zookeeper.set(dest_path, new_data, version=current_version)
    except BadVersionError:
        raise UnexpectedChangeError

    logger.info("Wrote data to Zookeeper.")
    return True


def main() -> None:
    arg_parser = argparse.ArgumentParser(description=sys.modules[__name__].__doc__)
    arg_parser.add_argument(
        "--debug", default=False, action="store_true", help="enable debug logging"
    )
    arg_parser.add_argument(
        "config_file", type=argparse.FileType("r"), help="path to a configuration file"
    )
    arg_parser.add_argument("source", type=argparse.FileType("rb"), help="file to upload")
    arg_parser.add_argument("dest", help="path in zookeeper")
    args = arg_parser.parse_args()

    if args.debug:
        level = logging.DEBUG
    else:
        level = logging.INFO
    logging.basicConfig(level=level, format="%(message)s")

    # quiet kazoo's verbose logs a bit
    logging.getLogger("kazoo").setLevel(logging.WARNING)

    parser = configparser.RawConfigParser(interpolation=EnvironmentInterpolation())
    parser.read_file(args.config_file)
    watcher_config = dict(parser.items("live-data"))

    secrets = secrets_store_from_config(watcher_config)
    zookeeper = zookeeper_client_from_config(secrets, watcher_config)
    zookeeper.start()
    try:
        write_file_to_zookeeper(zookeeper, args.source, args.dest)
    except WriterError as exc:
        print(exc, file=sys.stderr)
        sys.exit(1)
    finally:
        zookeeper.stop()


if __name__ == "__main__":
    main()



import unittest
import uuid

from io import BytesIO
from unittest import mock

try:
    from kazoo.client import KazooClient
    from kazoo.exceptions import NoNodeError
except ImportError:
    raise unittest.SkipTest("kazoo is not installed")

from baseplate.lib.live_data.writer import (
    NodeDoesNotExistError,
    UnexpectedChangeError,
    write_file_to_zookeeper,
)

from .. import get_endpoint_or_skip_container


zookeeper_endpoint = get_endpoint_or_skip_container("zookeeper", 2181)


# randomize the node path so simultaneously running test suites don't clobber one another
TEST_NODE_PATH = f"/writer-test-{uuid.uuid4()}"


class LiveDataWriterTests(unittest.TestCase):
    def setUp(self):
        self.zookeeper = KazooClient(hosts="%s:%d" % zookeeper_endpoint.address)
        self.zookeeper.start()

        try:
            self.zookeeper.delete(TEST_NODE_PATH)
        except NoNodeError:
            pass

        self.zookeeper.create(TEST_NODE_PATH, b"")

    def tearDown(self):
        self.zookeeper.stop()

    def test_exits_when_node_does_not_exist(self):
        input = BytesIO()

        with self.assertRaises(NodeDoesNotExistError):
            write_file_to_zookeeper(self.zookeeper, input, "/does_not_exist")

    def test_current_data_matches_new_data(self):
        data = b"data"
        self.zookeeper.set(TEST_NODE_PATH, data)

        input = BytesIO(data)
        did_write = write_file_to_zookeeper(self.zookeeper, input, TEST_NODE_PATH)
        self.assertFalse(did_write)

    def test_successful_set(self):
        self.assertEqual(self.zookeeper.get(TEST_NODE_PATH)[0], b"")

        input = BytesIO(b"new_data")
        did_write = write_file_to_zookeeper(self.zookeeper, input, TEST_NODE_PATH)
        self.assertTrue(did_write)

        self.assertEqual(self.zookeeper.get(TEST_NODE_PATH)[0], b"new_data")

    def test_version_changed(self):
        # this is a horrible hack. we want to test what happens if the contents
        # of the znode change between when write_file_to_zookeeper reads and
        # when it writes (so that the diff is trustworthy). this mock side
        # effect allows us to inject some behavior into the middle of the
        # function.
        mock_file = mock.Mock()

        def mock_read():
            self.zookeeper.set(TEST_NODE_PATH, b"I changed!")
            return b"contents of file"

        mock_file.read.side_effect = mock_read

        with self.assertRaises(UnexpectedChangeError):
            write_file_to_zookeeper(self.zookeeper, mock_file, TEST_NODE_PATH)



"""Daemon that fetches secrets from Vault and writes them to disk.

This daemon looks for configuration in an INI file passed on the command line.
The file should contain a section looking like:

    [secret-fetcher]
    vault.url = https://vault.example.com:8200/
    vault.role = my-server-role
    vault.auth_type = {aws,kubernetes}
    vault.mount_point = {aws-ec2,kubernetes}

    output.path = /var/local/secrets.json
    output.owner = www-data
    output.group = www-data
    output.mode = 0400

    secrets =
        secret/one,
        secret/two,
        secret/three,

    callback = scripts/my-transformer  # optional

where each secret is a path to look up in Vault. The daemon authenticates with
Vault as a role using a token obtained from an auth backend designated by `auth_type`.

Currently supported auth types:
    - aws: uses an AWS-signed instance identity document from the instance
    metadata API
    - kubernetes: uses a JWT mounted within a pod associated with a service account

Upon authenticating with this token, the Vault client then gets access based
upon the policies mapped to the role.

The secrets will be read from Vault and written to output.path as a JSON file
with the following structure:

    {
        "secrets": {
            "secret/one": {...},
            "secret/two": {...},
            "secret/three": {...}
        },
        "vault": {
            "token": "9da4241c-3460-11e7-84ac-0e9f9d32522f",
            "url": "https://vault.example.com:8200/"
        }
    }

The vault token and URL can be used for direct communication with Vault using
the server's authority. The file will be updated as secrets expire and need to
be refetched.

The `store` module in this package contains utilities for interacting with this
file from a running service.

Some applications require a specific format for their secrets. The `callback` option may
be provided if a different format is needed. This script will be invoked as a subprocess
with the JSON file as its first and only argument. This allows you to read in the secrets,
write to a new file in whatever format needed, and restart other services if necessary.

"""
import argparse
import configparser
import datetime
import json
import logging
import os
import posixpath
import subprocess
import time
import urllib.parse
import uuid

from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple

import requests

from baseplate import __version__ as baseplate_version
from baseplate.lib import config
from baseplate.server import EnvironmentInterpolation


logger = logging.getLogger(__name__)


K8S_SERVICE_ACCOUNT_TOKEN_FILE = "/var/run/secrets/kubernetes.io/serviceaccount/token"
NONCE_FILENAME = "/var/local/vault.nonce"
VAULT_TOKEN_PREFETCH_TIME = datetime.timedelta(seconds=60)
REAUTHENTICATION_ERROR_MESSAGE = """
Authenication failed! If this instance previously authenticated with a
different nonce, a vault operator may need to remove the instance ID from the
identity whitelist. See
https://www.vaultproject.io/docs/auth/aws.html#client-nonce
""".replace(
    "\n", " "
)


def fetch_instance_identity() -> str:
    """Retrieve the instance identity document from the metadata service.

    http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instance-identity-documents.html

    """
    logger.debug("Fetching identity.")
    resp = requests.get("http://169.254.169.254/latest/dynamic/instance-identity/pkcs7", timeout=5)
    resp.raise_for_status()
    return resp.text


def generate_nonce() -> str:
    """Return a string value suitable for use as a nonce."""
    logger.debug("Generating a new nonce.")
    return str(uuid.uuid4())


def load_nonce() -> Optional[str]:
    """Load the nonce from disk."""
    try:
        logger.debug("Loading nonce.")
        with open(NONCE_FILENAME, encoding="UTF-8") as f:
            return f.read()
    except OSError as exc:
        logger.debug("Nonce not found: %s.", exc)
        return None


def store_nonce(nonce: str) -> None:
    """Store the nonce to disk securely."""
    logger.debug("Storing nonce.")
    fd = os.open(NONCE_FILENAME, os.O_WRONLY | os.O_CREAT, 0o400)
    with os.fdopen(fd, "w") as f:
        f.write(nonce)


def ttl_to_time(ttl: int) -> datetime.datetime:
    """Return an absolute expiration time given a TTL."""
    return datetime.datetime.utcnow() + datetime.timedelta(seconds=ttl)


Authenticator = Callable[["VaultClientFactory"], Tuple[str, datetime.datetime]]


class VaultClientFactory:
    """Factory that makes authenticated clients."""

    def __init__(self, base_url: str, role: str, auth_type: Authenticator, mount_point: str):
        self.base_url = base_url
        self.role = role
        self.auth_type = auth_type
        self.mount_point = mount_point
        self.session = requests.Session()
        self.session.headers[
            "User-Agent"
        ] = f"baseplate.py-{self.__class__.__name__}/{baseplate_version}"
        self.client: Optional["VaultClient"] = None

    def _make_client(self) -> "VaultClient":
        """Obtain a client token from an auth backend and return a Vault client with it."""
        client_token, lease_duration = self.auth_type(self)

        return VaultClient(self.session, self.base_url, client_token, lease_duration)

    def _vault_kubernetes_auth(self) -> Tuple[str, datetime.datetime]:
        r"""Get a client token from Vault through the Kubernetes auth backend.

        This authenticates with Vault as a specified role using its
        Kubernetes auth backend. This involves sending Vault a JSON Web Token
        associated with a Kubernetes service account mounted at a well-known
        location within a running pod. Vault should be configured with a
        mapping binding roles to corresponding Kubernetes service accounts and
        namespaces along with appropriate policies. For example, a pod running
        in the `prod` namespace with the service account name `my-server`
        requires a Vault configuration created like so:

            vault write /auth/kubernetes/cluster-name/role/my-server-role \
                bound_service_account_names=my-server \
                bound_service_account_namespaces=prod \
                policies=my-servers-policies \
                max_ttl=4h

        See https://www.vaultproject.io/docs/auth/kubernetes.html for more info.

        """
        try:
            with open(K8S_SERVICE_ACCOUNT_TOKEN_FILE, encoding="UTF-8") as f:
                token = f.read()
        except OSError:
            logger.error(
                "Could not read Kubernetes token file '%s'", K8S_SERVICE_ACCOUNT_TOKEN_FILE
            )
            raise

        login_data = {"jwt": token, "role": self.role}

        logger.debug("Obtaining Vault token via kubernetes auth.")
        response = self.session.post(
            urllib.parse.urljoin(self.base_url, f"v1/auth/{self.mount_point}/login"),
            json=login_data,
            timeout=5,  # seconds
        )
        response.raise_for_status()
        auth = response.json()["auth"]
        return auth["client_token"], ttl_to_time(auth["lease_duration"])

    def _vault_aws_auth(self) -> Tuple[str, datetime.datetime]:
        r"""Get a client token from Vault through the AWS auth backend.

        This authenticates with Vault as a specified role using its AWS
        auth backend. This involves sending to Vault the AWS-signed instance
        identity document from the instance metadata API. Vault should have an
        appropriate role-mapping configured for the server so that appropriate
        policies can be applied to the returned token. For example, to
        authenticate any server with the `my-servers-iam-role` IAM Role:

            vault write /auth/aws-ec2/role/my-server-role \
                bound_iam_arn=arn:aws:iam::12341234:role/my-servers-iam-role \
                policies=my-servers-policies \
                max_ttl=4h

        To combat replay attacks where the identity document is snarfed and
        passed on by an interloper, during first login we store a
        Vault-generated nonce locally in a protected file. This nonce must be
        passed back to Vault on all successive login attempts.

        See https://www.vaultproject.io/docs/auth/aws.html for more info.

        """
        identity_document = fetch_instance_identity()
        nonce = load_nonce()
        if not nonce:
            # By generating our own nonce rather than relying on Vault to
            # generate one for us, we can avoid vault logging the nonce value
            # in plaintext audit logs.
            # https://www.vaultproject.io/docs/auth/aws.html#client-nonce
            nonce = generate_nonce()
            store_nonce(nonce)

        login_data = {"role": self.role, "pkcs7": identity_document, "nonce": nonce}

        logger.debug("Obtaining Vault token via aws auth.")
        response = self.session.post(
            urllib.parse.urljoin(self.base_url, f"v1/auth/{self.mount_point}/login"),
            json=login_data,
            timeout=5,  # seconds
        )
        if response.status_code == 400:
            logger.error(REAUTHENTICATION_ERROR_MESSAGE)
        response.raise_for_status()
        auth = response.json()["auth"]
        return auth["client_token"], ttl_to_time(auth["lease_duration"])

    @staticmethod
    def auth_types() -> Dict[str, Authenticator]:
        """Return a dict of the supported auth types and respective methods."""
        return {
            "aws": VaultClientFactory._vault_aws_auth,
            "kubernetes": VaultClientFactory._vault_kubernetes_auth,
        }

    def get_client(self) -> "VaultClient":
        """Get an authenticated client, reauthenticating if not cached."""
        if not self.client or self.client.is_about_to_expire:
            self.client = self._make_client()
        return self.client


class VaultClient:
    """An authenticated vault client.

    Use this as long as is_about_to_expire is False. Once that becomes True,
    get a new client from the factory.

    """

    def __init__(
        self,
        session: requests.Session,
        base_url: str,
        token: str,
        token_expiration: datetime.datetime,
    ):
        self.session = session
        self.base_url = base_url
        self.token = token
        self.token_expiration = token_expiration

    @property
    def is_about_to_expire(self) -> bool:
        """Return if the token is near expiration and in need of regeneration."""
        expiration = self.token_expiration - VAULT_TOKEN_PREFETCH_TIME
        return expiration < datetime.datetime.utcnow()

    def get_secret(self, secret_name: str) -> Tuple[Any, datetime.datetime]:
        """Get the value and expiration time of a named secret."""
        logger.debug("Fetching secret %r.", secret_name)
        try:
            response = self.session.get(
                urllib.parse.urljoin(self.base_url, posixpath.join("v1", secret_name)),
                headers={"X-Vault-Token": self.token},
                timeout=5,  # seconds
            )
            response.raise_for_status()
        except requests.HTTPError as e:
            logger.error("Vault client failed to get secret: %s: %s", secret_name, e)
            raise

        payload = response.json()
        return payload["data"], ttl_to_time(payload["lease_duration"])


def fetch_secrets(
    cfg: config.ConfigNamespace, client_factory: VaultClientFactory
) -> datetime.datetime:
    logger.info("Fetching secrets.")
    client = client_factory.get_client()
    secrets = {}
    soonest_expiration = client.token_expiration
    for secret_name in cfg.secrets:
        secrets[secret_name], expiration = client.get_secret(secret_name)
        soonest_expiration = min(soonest_expiration, expiration)

    with open(cfg.output.path + ".tmp", "w", encoding="UTF-8") as f:
        os.fchown(f.fileno(), cfg.output.owner, cfg.output.group)
        os.fchmod(f.fileno(), cfg.output.mode)

        json.dump(
            {
                "secrets": secrets,
                "vault": {"token": client.token, "url": cfg.vault.url},
                # this is here to allow an upgrade path. the fetcher should
                # be upgraded first followed by the application workers.
                "vault_token": client.token,
            },
            f,
            indent=2,
            sort_keys=True,
        )

    # swap out the file contents atomically
    os.rename(cfg.output.path + ".tmp", cfg.output.path)
    logger.info("Secrets fetched.")
    return soonest_expiration


def trigger_callback(
    callback: Optional[str], secrets_file: str, last_proc: Optional[subprocess.Popen] = None
) -> Optional[subprocess.Popen]:
    if callback:
        if last_proc and last_proc.poll() is None:
            logger.info("Previous callback process is still running. Skipping")
        else:
            return subprocess.Popen([callback, secrets_file])  # pylint: disable=R1732
    return last_proc


def main() -> None:
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        "config_file", type=argparse.FileType("r"), help="path to a configuration file"
    )
    arg_parser.add_argument(
        "--debug", default=False, action="store_true", help="enable debug logging"
    )
    arg_parser.add_argument(
        "--once",
        default=False,
        action="store_true",
        help="only run the fetcher once rather than as a daemon",
    )
    args = arg_parser.parse_args()

    if args.debug:
        level = logging.DEBUG
    else:
        level = logging.INFO

    logging.basicConfig(format="%(asctime)s:%(levelname)s:%(message)s", level=level)
    parser = configparser.RawConfigParser(interpolation=EnvironmentInterpolation())
    parser.read_file(args.config_file)
    fetcher_config = dict(parser.items("secret-fetcher"))

    cfg = config.parse_config(
        fetcher_config,
        {
            "vault": {
                "url": config.DefaultFromEnv(config.String, "BASEPLATE_DEFAULT_VAULT_URL"),
                "role": config.String,
                "auth_type": config.Optional(
                    config.OneOf(**VaultClientFactory.auth_types()),
                    default=VaultClientFactory.auth_types()["aws"],
                ),
                "mount_point": config.DefaultFromEnv(
                    config.String, "BASEPLATE_VAULT_MOUNT_POINT", fallback="aws-ec2"
                ),
            },
            "output": {
                "path": config.Optional(config.String, default="/var/local/secrets.json"),
                "owner": config.Optional(config.UnixUser, default=0),
                "group": config.Optional(config.UnixGroup, default=0),
                "mode": config.Optional(config.Integer(base=8), default=0o400),  # type: ignore
            },
            "secrets": config.Optional(config.TupleOf(config.String), default=[]),
            "callback": config.Optional(config.String),
        },
    )

    # pylint: disable=maybe-no-member
    client_factory = VaultClientFactory(
        cfg.vault.url, cfg.vault.role, cfg.vault.auth_type, cfg.vault.mount_point
    )

    if args.once:
        logger.info("Running secret fetcher once")
        fetch_secrets(cfg, client_factory)
        trigger_callback(cfg.callback, cfg.output.path)
    else:
        logger.info("Running secret fetcher as a daemon")
        last_proc = None
        while True:
            soonest_expiration = fetch_secrets(cfg, client_factory)
            last_proc = trigger_callback(cfg.callback, cfg.output.path, last_proc)
            time_til_expiration = soonest_expiration - datetime.datetime.utcnow()
            time_to_sleep = time_til_expiration - VAULT_TOKEN_PREFETCH_TIME
            time.sleep(max(int(time_to_sleep.total_seconds()), 1))


if __name__ == "__main__":
    main()



"""Internal helpers for the requests HTTP client library.

This stuff is not stable yet, so it's only for baseplate-internal use.

"""
import socket
import urllib.parse

from typing import Mapping
from typing import Optional

import requests.adapters
import urllib3.connectionpool

# the adapter code below is inspired by similar code in
# https://github.com/msabramo/requests-unixsocket/blob/master/requests_unixsocket/adapters.py
# https://github.com/docker/docker-py/blob/master/docker/unixconn/unixconn.py


class _UNIXConnection(urllib3.connectionpool.HTTPConnection):
    # pylint: disable=super-init-not-called
    def __init__(self, url: str):
        urllib3.connectionpool.HTTPConnection.__init__(self, "localhost")
        self.url = urllib.parse.urlparse(url)

    def connect(self) -> None:
        socket_path = urllib.parse.unquote(self.url.netloc)

        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.settimeout(1)
        self.sock.connect(socket_path)


class _UNIXConnectionPool(urllib3.connectionpool.HTTPConnectionPool):
    def __init__(self, url: str):
        super().__init__(host="localhost")
        self.url = url

    def _new_conn(self) -> _UNIXConnection:
        return _UNIXConnection(self.url)


class _UNIXAdapter(requests.adapters.HTTPAdapter):
    def get_connection(
        self, url: str, proxies: Optional[Mapping[str, str]] = None
    ) -> _UNIXConnectionPool:
        assert not proxies, "proxies are not supported"
        return _UNIXConnectionPool(url)


def add_unix_socket_support(session: requests.Session) -> None:
    """Add support to a Requests session for HTTP over UNIX domain sockets."""
    session.mount("http+unix://", _UNIXAdapter())



import socket
import tempfile
import unittest

from unittest.mock import patch

from baseplate.lib import config


class StringTests(unittest.TestCase):
    def test_parse_string(self):
        result = config.String("whatever")
        self.assertEqual(result, "whatever")

    def test_empty_string_not_ok(self):
        with self.assertRaises(ValueError):
            config.String("")


class IntegerTests(unittest.TestCase):
    def test_parse_integer_valid(self):
        result = config.Integer("337")
        self.assertEqual(result, 337)

    def test_parse_integer_invalid(self):
        with self.assertRaises(ValueError):
            config.Integer("")
        with self.assertRaises(ValueError):
            config.Integer("illegal")

    def test_parse_integer_actually_float(self):
        with self.assertRaises(ValueError):
            config.Integer("1.2")

    def test_non_decimal(self):
        result = config.Integer(base=8)("0600")
        self.assertEqual(result, 384)


class FloatTests(unittest.TestCase):
    def test_parse_float_valid(self):
        result = config.Float("1.2")
        self.assertEqual(result, 1.2)

    def test_parse_float_invalid(self):
        with self.assertRaises(ValueError):
            config.Float("")
        with self.assertRaises(ValueError):
            config.Float("sdklfj")


class BooleanTests(unittest.TestCase):
    def test_boolean_valid(self):
        self.assertEqual(config.Boolean("true"), True)
        self.assertEqual(config.Boolean("True"), True)
        self.assertEqual(config.Boolean("false"), False)
        self.assertEqual(config.Boolean("False"), False)

    def test_boolean_invalid(self):
        with self.assertRaises(ValueError):
            config.Boolean("")
        with self.assertRaises(ValueError):
            config.Boolean("wrong")


class EndpointTests(unittest.TestCase):
    def test_endpoint_empty(self):
        with self.assertRaises(ValueError):
            config.Endpoint("")

    def test_endpoint_inet(self):
        result = config.Endpoint("localhost:1234")
        self.assertEqual(result.family, socket.AF_INET)
        self.assertEqual(result.address, ("localhost", 1234))

    def test_endpoint_inet_invalid(self):
        with self.assertRaises(ValueError):
            config.Endpoint("localhost")

    def test_endpoint_unix(self):
        result = config.Endpoint("/this/is/a/path.sock")
        self.assertEqual(result.family, socket.AF_UNIX)
        self.assertEqual(result.address, "/this/is/a/path.sock")


class Base64Tests(unittest.TestCase):
    def test_invalid(self):
        with self.assertRaises(ValueError):
            config.Base64("")

        with self.assertRaises(ValueError):
            config.Base64("dGvzdAo")  # missing padding

    def test_valid(self):
        result = config.Base64("aHVudGVyMg==")
        self.assertEqual(result, b"hunter2")


class FileTests(unittest.TestCase):
    def setUp(self):
        self.tempfile = tempfile.NamedTemporaryFile()
        self.tempfile.write(b"test")
        self.tempfile.flush()

    def tearDown(self):
        self.tempfile.close()

    def test_no_file(self):
        file_opener = config.File()
        with self.assertRaises(ValueError):
            file_opener("/tmp/does_not_exist")

    def test_read_file(self):
        file_opener = config.File()
        the_file = file_opener(self.tempfile.name)
        self.assertEqual(the_file.read(), "test")

    def test_write_file(self):
        file_opener = config.File(mode="w")
        the_file = file_opener(self.tempfile.name)
        the_file.write("cool")
        the_file.close()

        with open(self.tempfile.name) as f:
            self.assertEqual(f.read(), "cool")


class TimespanTests(unittest.TestCase):
    def test_timespan_invalid(self):
        with self.assertRaises(ValueError):
            config.Timespan("")

        with self.assertRaises(ValueError):
            config.Timespan("a b")

        with self.assertRaises(ValueError):
            config.Timespan("10 florgles")

        with self.assertRaises(ValueError):
            config.Timespan("a b c")

        with self.assertRaises(ValueError):
            config.Timespan("3.2 hours")

    def test_timespan(self):
        result = config.Timespan("30 milliseconds")
        self.assertAlmostEqual(result.total_seconds(), 0.03)

        result = config.Timespan("1 second")
        self.assertEqual(result.total_seconds(), 1)

        result = config.Timespan("2 seconds")
        self.assertEqual(result.total_seconds(), 2)

        result = config.Timespan("30 minutes")
        self.assertEqual(result.total_seconds(), 1800)

        result = config.Timespan("2 hours")
        self.assertEqual(result.total_seconds(), 7200)

        result = config.Timespan("1 day")
        self.assertEqual(result.total_seconds(), 86400)


class TimespanWithLegacyFallbackTests(unittest.TestCase):
    def test_fallback(self):
        result = config.TimespanWithLegacyFallback("30 minutes")
        self.assertEqual(result.total_seconds(), 1800)

        result = config.TimespanWithLegacyFallback("92")
        self.assertEqual(result.total_seconds(), 92)


class PercentTests(unittest.TestCase):
    def test_percentage(self):
        self.assertAlmostEqual(config.Percent("37.2%"), 0.372)
        self.assertAlmostEqual(config.Percent("100%"), 1.0)

    def test_invalid_percentage(self):
        with self.assertRaises(ValueError):
            config.Percent("9")

        with self.assertRaises(ValueError):
            config.Percent("-10%")

        with self.assertRaises(ValueError):
            config.Percent("120%")

        with self.assertRaises(ValueError):
            config.Percent("30%%%%")


class UnixUserTests(unittest.TestCase):
    def test_valid_user(self):
        result = config.UnixUser("root")
        self.assertEqual(result, 0)

    def test_invalid_user(self):
        with self.assertRaises(ValueError):
            config.UnixUser("fhqwhgads")

    def test_uid_fallback(self):
        result = config.UnixUser("1000")
        self.assertEqual(result, 1000)


class UnixGroupTests(unittest.TestCase):
    def test_valid_group(self):
        result = config.UnixGroup("root")
        self.assertEqual(result, 0)

    def test_invalid_group(self):
        with self.assertRaises(ValueError):
            config.UnixGroup("fhqwhgads")

    def test_gid_fallback(self):
        result = config.UnixGroup("1000")
        self.assertEqual(result, 1000)


class OneOfTests(unittest.TestCase):
    def test_oneof_valid(self):
        parser = config.OneOf(ONE=1, TWO=2, THREE=3)
        self.assertEqual(parser("ONE"), 1)
        self.assertEqual(parser("TWO"), 2)
        self.assertEqual(parser("THREE"), 3)

    def test_oneof_invalid(self):
        parser = config.OneOf(ONE=1, TWO=2, THREE=3)
        with self.assertRaises(ValueError):
            parser("")
        with self.assertRaises(ValueError):
            parser("FOUR")


class TupleTests(unittest.TestCase):
    def test_tupleof_valid(self):
        parser = config.TupleOf(config.Integer)
        self.assertEqual(parser("1,2,3"), [1, 2, 3])
        self.assertEqual(parser("4, 5, 6"), [4, 5, 6])

    def test_tupleof_invalid(self):
        parser = config.TupleOf(config.Integer)

        with self.assertRaises(ValueError):
            parser("")

        with self.assertRaises(ValueError):
            parser("a, b")


@patch.dict("os.environ", {"BASEPLATE_DEFAULT_VALUE": "default", "NOT_PROVIDED": ""})
class DefaultFromEnvTests(unittest.TestCase):
    def test_use_default_from_env(self):
        parser = config.DefaultFromEnv(config.String, "BASEPLATE_DEFAULT_VALUE")
        self.assertEqual(parser(""), "default")

    def test_empty_default(self):
        parser = config.DefaultFromEnv(config.String, "NOT_PROVIDED")
        self.assertEqual(parser("foo"), "foo")

    def test_use_provided(self):
        parser = config.DefaultFromEnv(config.String, "BASEPALTE_DEFAULT_VALUE")
        self.assertEqual(parser("foo"), "foo")

    def test_provide_none(self):
        parser = config.DefaultFromEnv(config.String, "NOT_PROVIDED")
        self.assertRaises(ValueError, parser, "")

    def test_fallback(self):
        fallback_value = 5
        parser = config.DefaultFromEnv(config.Integer, "NOT_PROVIDED", fallback_value)
        self.assertEqual(parser(""), fallback_value)


class OptionalTests(unittest.TestCase):
    def test_optional_exists(self):
        parser = config.Optional(config.Integer)
        self.assertEqual(parser("33"), 33)

    def test_optional_default(self):
        parser = config.Optional(config.Integer)
        self.assertEqual(parser(""), None)

    def test_optional_invalid(self):
        parser = config.Optional(config.Integer)
        with self.assertRaises(ValueError):
            parser("asdf")


class FallbackTests(unittest.TestCase):
    def test_primary_option_works(self):
        parser = config.Fallback(config.Percent, config.Float)
        self.assertAlmostEqual(parser("33%"), 0.33)

    def test_fallback_option_works(self):
        parser = config.Fallback(config.Percent, config.Float)
        self.assertAlmostEqual(parser(".44"), 0.44)


class DictOfTests(unittest.TestCase):
    def test_empty(self):
        parser = config.DictOf(config.Integer)
        result = parser.parse("my_key", {"not_related": "a"})
        self.assertEqual(result, {})

    def test_scalar_children(self):
        parser = config.DictOf(config.Integer)
        result = parser.parse("my_key", {"my_key.a": "1", "my_key.b": "2"})
        self.assertEqual(result, {"a": 1, "b": 2})

    def test_vector_children(self):
        parser = config.DictOf({"a": config.Integer, "b": config.String})
        result = parser.parse(
            "my_key",
            {
                "my_key.first.a": "1",
                "my_key.first.b": "test",
                "my_key.first.c": "ignored",
                "my_key.second.a": "2",
                "my_key.second.b": "test",
            },
        )
        self.assertEqual(result, {"first": {"a": 1, "b": "test"}, "second": {"a": 2, "b": "test"}})

    def test_root_level(self):
        parser = config.DictOf({"a": config.Integer, "b": config.String})
        result = parser.parse(
            "",
            {
                "first.a": "1",
                "first.b": "test",
                "first.c": "ignored",
                "second.a": "2",
                "second.b": "test",
            },
        )
        self.assertEqual(result, {"first": {"a": 1, "b": "test"}, "second": {"a": 2, "b": "test"}})


class TestParseConfig(unittest.TestCase):
    def setUp(self):
        self.config = {
            "simple": "oink",
            "foo.bar": "33",
            "foo.baz": "a cool guy",
            "noo.bar": "",
            "deep.so.deep": "very",
        }

    def test_simple_config(self):
        result = config.parse_config(
            self.config,
            {
                "simple": config.String,
                "foo": {"bar": config.Integer},
                "noo": {"bar": config.Optional(config.String, default="")},
                "deep": {"so": {"deep": config.String}},
            },
        )

        self.assertEqual(result.simple, "oink")
        self.assertEqual(result.foo.bar, 33)
        self.assertEqual(result.noo.bar, "")
        self.assertEqual(result.deep.so.deep, "very")

    def test_missing_key(self):
        with self.assertRaises(config.ConfigurationError):
            config.parse_config(self.config, {"foo": {"not_here": config.Integer}})

    def test_bad_value(self):
        with self.assertRaises(config.ConfigurationError):
            config.parse_config(self.config, {"foo": {"baz": config.Integer}})

    def test_dot_in_key(self):
        with self.assertRaises(AssertionError):
            config.parse_config(self.config, {"foo.bar": {}})

    def test_spec_contains_invalid_object(self):
        with self.assertRaises(AssertionError):
            config.parse_config(self.config, {"tree_people": 37})

    def test_subparsers(self):
        result = config.parse_config(self.config, {"foo": config.DictOf(config.String)})
        self.assertEqual(result, {"foo": {"bar": "33", "baz": "a cool guy"}})



import os
import unittest

from unittest import mock

import fakeredis
import pytest

from prometheus_client import REGISTRY
from rediscluster.exceptions import RedisClusterException

from baseplate.clients.redis_cluster import ACTIVE_REQUESTS
from baseplate.clients.redis_cluster import cluster_pool_from_config
from baseplate.clients.redis_cluster import HotKeyTracker
from baseplate.clients.redis_cluster import LATENCY_SECONDS
from baseplate.clients.redis_cluster import MonitoredRedisClusterConnection
from baseplate.clients.redis_cluster import REQUESTS_TOTAL


class DummyConnection(object):
    description_format = "DummyConnection<>"

    def __init__(self, host="localhost", port=7000, socket_timeout=None, **kwargs):
        self.kwargs = kwargs
        self.pid = os.getpid()
        self.host = host
        self.port = port
        self.socket_timeout = socket_timeout

    def get_connection(self):
        pass

    def connect(self):
        pass

    def can_read(self):
        return False

    def send_command(self, command):
        pass

    def read_response(self):
        # Must return same number as test_pipeline_instrumentation calls
        return ["OK", "OK"]

    def pack_commands(self, *args):
        pass

    def send_packed_command(self, *args):
        pass

    def disconnect(self):
        pass


class TestMonitoredRedisConnection:
    def setup(self):
        ACTIVE_REQUESTS.clear()
        REQUESTS_TOTAL.clear()
        LATENCY_SECONDS.clear()

    @pytest.fixture
    def app_config(self):
        yield {
            "redis.url": "redis://localhost:1234/0",
        }

    @pytest.fixture
    def expected_labels(self):
        yield {
            "redis_client_name": "test_client",
            "redis_type": "cluster",
            "redis_command": "some_command",
            "redis_database": "0",
        }

    @pytest.fixture
    def connection(self):
        yield {
            "connection_class": DummyConnection,
        }

    @pytest.fixture
    def connection_pool(self, app_config, connection):
        return cluster_pool_from_config(
            app_config=app_config,
            prefix="redis.",
            init_slot_cache=False,
            startup_nodes=[
                {
                    "host": "127.0.0.1",
                    "port": 7000,
                }
            ],
            **connection,
        )

    @pytest.fixture
    def context(self):
        yield mock.MagicMock()

    @pytest.fixture
    def span(self):
        yield mock.MagicMock()

    @pytest.fixture
    @mock.patch("rediscluster.RedisCluster", new=mock.MagicMock())
    def monitored_redis_connection(self, span, connection_pool):
        return MonitoredRedisClusterConnection(
            "redis_context_name", span, connection_pool, redis_client_name="test_client"
        )

    # NOTE: a successful execute_command() call is difficult to mock
    def test_execute_command_exc_redis_err(
        self, monitored_redis_connection, expected_labels, app_config
    ):
        with pytest.raises(RedisClusterException):
            monitored_redis_connection.execute_command("some_command")
        assert REGISTRY.get_sample_value(f"{ACTIVE_REQUESTS._name}", expected_labels) == 0
        expected_labels["redis_success"] = "false"
        assert (
            REGISTRY.get_sample_value(
                f"{LATENCY_SECONDS._name}_bucket", {**expected_labels, **{"le": "+Inf"}}
            )
            == 1
        )
        assert REGISTRY.get_sample_value(f"{REQUESTS_TOTAL._name}_total", expected_labels) == 1

    def test_pipeline_instrumentation(self, monitored_redis_connection, expected_labels):
        active_labels = {**expected_labels, "redis_command": "pipeline"}
        mock_manager = mock.Mock()
        with mock.patch.object(
            ACTIVE_REQUESTS.labels(**active_labels),
            "inc",
            wraps=ACTIVE_REQUESTS.labels(**active_labels).inc,
        ) as active_inc_spy_method:
            mock_manager.attach_mock(active_inc_spy_method, "inc")
            with mock.patch.object(
                ACTIVE_REQUESTS.labels(**active_labels),
                "dec",
                wraps=ACTIVE_REQUESTS.labels(**active_labels).dec,
            ) as active_dec_spy_method:
                mock_manager.attach_mock(active_dec_spy_method, "dec")

                # This KeyError is the same problem as the RedisClusterException in `test_execute_command_exc_redis_err` above
                with pytest.raises(KeyError):
                    monitored_redis_connection.pipeline("test").set("hello", 42).set(
                        "goodbye", 23
                    ).execute()
                labels = {**active_labels, "redis_success": "false"}
                assert (
                    REGISTRY.get_sample_value(f"{REQUESTS_TOTAL._name}_total", labels) == 1.0
                ), "Unexpected value for REQUESTS_TOTAL metric. Expected one 'pipeline' command"
                assert (
                    REGISTRY.get_sample_value(
                        f"{LATENCY_SECONDS._name}_bucket", {**labels, "le": "+Inf"}
                    )
                    == 1.0
                ), "Expected one 'pipeline' latency request"
                assert mock_manager.mock_calls == [
                    mock.call.inc(),
                    mock.call.dec(),
                ], "Instrumentation should increment and then decrement active requests exactly once"
                print(list(REGISTRY.collect()))
                assert (
                    REGISTRY.get_sample_value(ACTIVE_REQUESTS._name, active_labels) == 0.0
                ), "Should have 0 (and not None) active requests"


class HotKeyTrackerTests(unittest.TestCase):
    def setUp(self):
        self.rc = fakeredis.FakeStrictRedis()

    def test_increment_reads_once(self):
        tracker = HotKeyTracker(self.rc, 1, 1)
        tracker.increment_keys_read_counter(["foo"], ignore_errors=False)
        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-reads", "-inf", "+inf", withscores=True
            ),
            [(b"foo", float(1))],
        )

    def test_increment_several_reads(self):
        tracker = HotKeyTracker(self.rc, 1, 1)
        for _ in range(5):
            tracker.increment_keys_read_counter(["foo"], ignore_errors=False)

        tracker.increment_keys_read_counter(["bar"], ignore_errors=False)

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-reads", "-inf", "+inf", withscores=True
            ),
            [(b"bar", float(1)), (b"foo", float(5))],
        )

    def test_reads_disabled_tracking(self):
        tracker = HotKeyTracker(self.rc, 0, 0)
        for _ in range(5):
            tracker.maybe_track_key_usage(["GET", "foo"])

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-reads", "-inf", "+inf", withscores=True
            ),
            [],
        )

    def test_reads_enabled_tracking(self):
        tracker = HotKeyTracker(self.rc, 1, 1)
        for _ in range(5):
            tracker.maybe_track_key_usage(["GET", "foo"])

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-reads", "-inf", "+inf", withscores=True
            ),
            [(b"foo", float(5))],
        )

    def test_writes_enabled_tracking(self):
        tracker = HotKeyTracker(self.rc, 1, 1)
        for _ in range(5):
            tracker.maybe_track_key_usage(["SET", "foo", "bar"])

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-writes", "-inf", "+inf", withscores=True
            ),
            [(b"foo", float(5))],
        )

    def test_writes_disabled_tracking(self):
        tracker = HotKeyTracker(self.rc, 0, 0)
        for _ in range(5):
            tracker.maybe_track_key_usage(["SET", "foo", "bar"])

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-writes", "-inf", "+inf", withscores=True
            ),
            [],
        )

    def test_write_multikey_commands(self):
        tracker = HotKeyTracker(self.rc, 1, 1)

        tracker.maybe_track_key_usage(["DEL", "foo", "bar"])

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-writes", "-inf", "+inf", withscores=True
            ),
            [(b"bar", float(1)), (b"foo", float(1))],
        )

    def test_write_batchkey_commands(self):
        tracker = HotKeyTracker(self.rc, 1, 1)

        tracker.maybe_track_key_usage(["MSET", "foo", "bar", "baz", "wednesday"])

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-writes", "-inf", "+inf", withscores=True
            ),
            [(b"baz", float(1)), (b"foo", float(1))],
        )



import glob
import os
import subprocess

from distutils.command.build_py import build_py
from distutils.core import Command

import baseplate.thrift


class BuildThriftCommand(Command):
    description = "Generate Python code from Thrift IDL."
    user_options = [("build-base=", "b", "base directory for build library")]

    def initialize_options(self) -> None:
        self.build_base = None

    def finalize_options(self) -> None:
        self.set_undefined_options("build", ("build_base", "build_base"))

    def run(self) -> None:
        if self.dry_run:  # type: ignore
            return

        temp_dir = os.path.join(self.build_base, "thrift")  # type: ignore
        self.mkpath(temp_dir)

        for package in self.distribution.packages:  # type: ignore
            package_dir = os.path.join(*package.split("."))

            for thriftfile in glob.glob(os.path.join(package_dir, "*.thrift")):
                subprocess.check_call(
                    [
                        "thrift",
                        "-strict",
                        "-gen",
                        "py:slots",
                        "-out",
                        temp_dir,
                        "-I",
                        os.path.dirname(baseplate.thrift.__file__),
                        thriftfile,
                    ]
                )

                module_name = os.path.splitext(os.path.basename(thriftfile))[0]
                input_package = os.path.join(temp_dir, module_name)

                generated_module = "_".join((module_name, "thrift"))
                output_package = os.path.join(package_dir, generated_module)

                self.copy_tree(input_package, output_package)

                # rewrite "from thriftname" to "from package.thriftname_thrift"
                full_package_name = f"{package}.{generated_module}"
                for remote in glob.glob(os.path.join(output_package, "*-remote")):
                    with open(remote) as f:
                        lines = f.readlines()

                    with open(remote, "w") as f:
                        for line in lines:
                            prefix = "from " + module_name
                            if line.startswith(prefix):
                                if not line.startswith("from " + full_package_name):
                                    f.write("from " + full_package_name + line[len(prefix) :])
                                    continue
                            f.write(line)


class ThriftBuildPyCommand(build_py):
    def run(self) -> None:
        self.run_command("build_thrift")
        build_py.run(self)



"""A Thrift client connection pool.

.. note:: See :py:class:`baseplate.clients.thrift.ThriftContextFactory` for
    a convenient way to integrate the pool with your application.

The pool lazily creates connections and maintains them in a pool. Individual
connections have a maximum lifetime, after which they will be recycled.

A basic example of usage::

    pool = thrift_pool_from_config(app_config, "example_service.")
    with pool.connection() as protocol:
        client = ExampleService.Client(protocol)
        client.do_example_thing()

"""
import contextlib
import logging
import queue
import socket
import time

from typing import Any
from typing import Generator
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING

from thrift.protocol import THeaderProtocol
from thrift.protocol.TProtocol import TProtocolBase
from thrift.protocol.TProtocol import TProtocolException
from thrift.protocol.TProtocol import TProtocolFactory
from thrift.Thrift import TApplicationException
from thrift.Thrift import TException
from thrift.transport.TSocket import TSocket
from thrift.transport.TTransport import TTransportException

from baseplate.lib import config
from baseplate.lib.retry import RetryPolicy


logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    ProtocolPool = Type[queue.Queue[TProtocolBase]]  # pylint: disable=unsubscriptable-object
else:
    ProtocolPool = queue.Queue


def _make_transport(endpoint: config.EndpointConfiguration) -> TSocket:
    if endpoint.family == socket.AF_INET:
        trans = TSocket(*endpoint.address)
    elif endpoint.family == socket.AF_UNIX:
        trans = TSocket(unix_socket=endpoint.address)
    else:
        raise Exception(f"unsupported endpoint family {endpoint.family!r}")

    return trans


def thrift_pool_from_config(
    app_config: config.RawConfig, prefix: str, **kwargs: Any
) -> "ThriftConnectionPool":
    """Make a ThriftConnectionPool from a configuration dictionary.

    The keys useful to :py:func:`thrift_pool_from_config` should be prefixed,
    e.g.  ``example_service.endpoint`` etc. The ``prefix`` argument specifies
    the prefix used to filter keys.  Each key is mapped to a corresponding
    keyword argument on the :py:class:`ThriftConnectionPool` constructor.  Any
    keyword arguments given to this function will be also be passed through to
    the constructor. Keyword arguments take precedence over the configuration
    file.

    Supported keys:

    * ``endpoint`` (required): A ``host:port`` pair, e.g. ``localhost:2014``,
        where the Thrift server can be found.
    * ``fifo_queue``: True will enable use of a FIFO queue instead of the default LIFO queue.
    * ``size``: The size of the connection pool.
    * ``max_age``: The oldest a connection can be before it's recycled and
        replaced with a new one. Written as a
        :py:func:`~baseplate.lib.config.Timespan` e.g. ``1 minute``.
    * ``timeout``: The maximum amount of time a connection attempt or RPC call
        can take before a TimeoutError is raised.
        (:py:func:`~baseplate.lib.config.Timespan`)
    * ``max_connection_attempts``: The maximum number of times the pool will attempt to
        open a connection.

    .. versionchanged:: 1.2
        ``max_retries`` was renamed ``max_connection_attempts``.

    """
    assert prefix.endswith(".")
    parser = config.SpecParser(
        {
            "endpoint": config.Endpoint,
            "fifo_queue": config.Optional(config.Boolean, default=False),
            "size": config.Optional(config.Integer, default=10),
            "max_age": config.Optional(config.Timespan, default=config.Timespan("1 minute")),
            "timeout": config.Optional(config.Timespan, default=config.Timespan("1 second")),
            "max_connection_attempts": config.Optional(config.Integer),
            "max_retries": config.Optional(config.Integer),
        }
    )
    options = parser.parse(prefix[:-1], app_config)

    if options.fifo_queue:
        kwargs.setdefault("queue_cls", queue.Queue)
    if options.size is not None:
        kwargs.setdefault("size", options.size)
    if options.max_age is not None:
        kwargs.setdefault("max_age", options.max_age.total_seconds())
    if options.timeout is not None:
        kwargs.setdefault("timeout", options.timeout.total_seconds())
    if options.max_connection_attempts is not None:
        kwargs.setdefault("max_connection_attempts", options.max_connection_attempts)
    if options.max_retries is not None:
        raise Exception("max_retries was renamed to max_connection_attempts")

    return ThriftConnectionPool(endpoint=options.endpoint, **kwargs)


_DEFAULT_PROTOCOL_FACTORY = THeaderProtocol.THeaderProtocolFactory(
    default_protocol=THeaderProtocol.THeaderSubprotocolID.COMPACT,
)


class ThriftConnectionPool:
    """A pool that maintains a queue of open Thrift connections.

    :param endpoint: The remote address
        of the Thrift service.
    :param size: The maximum number of connections that can be open
        before new attempts to open block.
    :param max_age: The maximum number of seconds a connection should be
        kept alive. Connections older than this will be reaped.
    :param timeout: The maximum number of seconds a connection attempt or
        RPC call can take before a TimeoutError is raised.
    :param max_connection_attempts: The maximum number of times the pool will attempt
        to open a connection.
    :param protocol_factory: The factory to use for creating protocols from
        transports. This is useful for talking to services that don't support
        THeaderProtocol.
    :param queue_cls: A stdlib compatible queue class.

    All exceptions raised by this class derive from
    :py:exc:`~thrift.transport.TTransport.TTransportException`.

    .. versionchanged:: 1.2
        ``max_retries`` was renamed ``max_connection_attempts``.

    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        endpoint: config.EndpointConfiguration,
        size: int = 10,
        max_age: int = 120,
        timeout: float = 1,
        max_connection_attempts: int = 3,
        protocol_factory: TProtocolFactory = _DEFAULT_PROTOCOL_FACTORY,
        queue_cls: ProtocolPool = queue.LifoQueue,
    ):
        self.endpoint = endpoint
        self.max_age = max_age
        self.retry_policy = RetryPolicy.new(attempts=max_connection_attempts)
        self.timeout = timeout
        self.protocol_factory = protocol_factory

        self.size = size

        self.pool = queue_cls()
        for _ in range(size):
            self.pool.put(None)

    def _get_from_pool(self) -> Optional[TProtocolBase]:
        try:
            return self.pool.get(block=True, timeout=self.timeout)
        except queue.Empty:
            raise TTransportException(
                type=TTransportException.NOT_OPEN, message="timed out waiting for a connection slot"
            )

    def _create_connection(self) -> TProtocolBase:
        for _ in self.retry_policy:
            trans = _make_transport(self.endpoint)
            trans.setTimeout(self.timeout * 1000.0)
            prot = self.protocol_factory.getProtocol(trans)

            try:
                prot.trans.open()
            except TTransportException as exc:
                logger.info("Failed to connect to %r: %s", self.endpoint, exc)
                continue

            prot.baseplate_birthdate = time.time()

            return prot

        raise TTransportException(
            type=TTransportException.NOT_OPEN,
            message="giving up after multiple attempts to connect",
        )

    def _is_stale(self, prot: TProtocolBase) -> bool:
        if not prot.trans.isOpen() or time.time() - prot.baseplate_birthdate > self.max_age:
            prot.trans.close()
            return True
        return False

    def _release(self, prot: Optional[TProtocolBase]) -> None:
        if prot and prot.trans.isOpen():
            self.pool.put(prot)
        else:
            self.pool.put(None)

    @contextlib.contextmanager
    def connection(self) -> Generator[TProtocolBase, None, None]:
        """Acquire a connection from the pool.

        This method is to be used with a context manager. It returns a
        connection from the pool, or blocks up to :attr:`timeout` seconds
        waiting for one if the pool is full and all connections are in use.

        When the context is exited, the connection is returned to the pool.
        However, if it was exited via an unexpected Thrift exception, the
        connection is closed instead because the state of the connection is
        unknown.

        """
        prot = self._get_from_pool()

        try:
            if not prot or self._is_stale(prot):
                prot = self._create_connection()

            try:
                yield prot
            except socket.timeout:
                # thrift doesn't re-wrap socket timeout errors appropriately so
                # we'll do it here for a saner exception hierarchy
                raise TTransportException(
                    type=TTransportException.TIMED_OUT, message="timed out interacting with socket"
                )
            except OSError as exc:
                raise TTransportException(type=TTransportException.UNKNOWN, message=str(exc))
        except (TApplicationException, TProtocolException, TTransportException):
            # these exceptions usually indicate something low-level went wrong,
            # so it's safest to just close this connection because we don't
            # know what state it's in.
            if prot:
                prot.trans.close()
            raise
        except TException:
            # the only other TException-derived errors are application level
            # (expected) errors which should be safe for the connection.
            # don't close the transport here!
            raise
        except:  # noqa: E722
            # anything else coming out of thrift usually means parsing failed
            # or something nastier. we'll just play it safe and close the
            # connection.
            if prot:
                prot.trans.close()
            raise
        finally:
            self._release(prot)

    @property
    def checkedout(self) -> int:
        return self.size - self.pool.qsize()



import io
import json
import os

from unittest import mock

import botocore.session
import pytest

from botocore.response import StreamingBody
from botocore.stub import Stubber
from moto import mock_s3

from baseplate.sidecars.live_data_watcher import _load_from_s3
from baseplate.sidecars.live_data_watcher import _parse_loader_type
from baseplate.sidecars.live_data_watcher import LoaderException
from baseplate.sidecars.live_data_watcher import LoaderType
from baseplate.sidecars.live_data_watcher import NodeWatcher


@pytest.fixture()
def start_mock_s3():
    with mock_s3():
        yield


@pytest.fixture()
def s3_stub(start_mock_s3):
    s3 = botocore.session.get_session().create_client("s3")
    with Stubber(s3) as stubber:
        with mock.patch("baseplate.sidecars.live_data_watcher.boto3.client") as client:
            client.return_value = s3
            yield stubber


@pytest.mark.parametrize(
    "data,return_value",
    (
        # Non-obviously-PASSTHROUGH configs
        (json.dumps({"live_data_watcher_load_type": "S3"}).encode("utf-8"), LoaderType.S3),
        (
            json.dumps({"live_data_watcher_load_type": "NOTS3"}).encode("utf-8"),
            LoaderType.PASSTHROUGH,
        ),
        # Everything else should be PASSTHROUGH
        (b"", LoaderType.PASSTHROUGH),
        (json.dumps({"key": "value"}).encode("utf-8"), LoaderType.PASSTHROUGH),
        (json.dumps([]).encode("utf-8"), LoaderType.PASSTHROUGH),
        (json.dumps(1234).encode("utf-8"), LoaderType.PASSTHROUGH),
        (json.dumps(None).encode("utf-8"), LoaderType.PASSTHROUGH),
        (b"hey{look}at--this//ugly**string@of(ugly^^chars", LoaderType.PASSTHROUGH),
    ),
)
def test_parse_loader_type(data, return_value):
    assert _parse_loader_type(data) == return_value


@pytest.mark.parametrize(
    "missing_config",
    (
        "region_name",
        "bucket_name",
        "file_key",
        "sse_key",
    ),
)
def test_load_from_s3_missing_config(start_mock_s3, missing_config):
    data = {
        "region_name": "some_region",
        "bucket_name": "some_bucket",
        "file_key": "some_file",
        "sse_key": "some_key",
    }
    data.pop(missing_config)

    with pytest.raises(LoaderException):
        assert _load_from_s3(json.dumps(data).encode("utf-8"))


@pytest.mark.parametrize(
    "anon",
    (
        True,
        False,
    ),
)
def test_successful_load_from_s3(s3_stub, anon):
    contents = b"my-test-contents"
    bucket_name = "my-test-bucket"
    file_key = "my-test-key"
    sse_key = "key"

    s3_stub.add_response(
        "get_object",
        service_response={
            "Body": StreamingBody(io.BytesIO(contents), len(contents)),
        },
        expected_params={
            "Bucket": bucket_name,
            "Key": file_key,
            "SSECustomerKey": sse_key,
            "SSECustomerAlgorithm": "AES256",
        },
    )

    data = _load_from_s3(
        json.dumps(
            {
                "region_name": "us-east-1",
                "bucket_name": bucket_name,
                "file_key": file_key,
                "sse_key": sse_key,
                "anon": anon,
            }
        ).encode("utf-8")
    )

    assert data == contents


def test_unsuccessful_load_from_s3_client_error(s3_stub):
    s3_stub.add_client_error("get_object")

    with pytest.raises(LoaderException):
        _load_from_s3(
            json.dumps(
                {
                    "region_name": "us-east-1",
                    "bucket_name": "my-test-bucket",
                    "file_key": "my-object-key",
                    "sse_key": "my-sse-key",
                }
            ).encode("utf-8")
        )


@pytest.mark.parametrize(
    "load_type,should_call_s3",
    [(LoaderType.S3, True)] + [(lt, False) for lt in LoaderType if lt != LoaderType.S3],
)
def test_loader_type_calls(tmp_path, load_type, should_call_s3):
    loader_path = "baseplate.sidecars.live_data_watcher._load_from_s3"
    parse_path = "baseplate.sidecars.live_data_watcher._parse_loader_type"
    with mock.patch(loader_path) as loader, mock.patch(parse_path) as parser:
        parser.return_value = load_type
        loader.return_value = b""
        watcher = NodeWatcher(str(tmp_path / "tmp"), os.getuid(), os.getgid(), 777)
        watcher.on_change(b"", None)

        assert loader.called == should_call_s3



"""Check health of a baseplate service on localhost."""
import argparse
import socket
import sys
import typing
import urllib.parse

import requests

from baseplate.lib._requests import add_unix_socket_support
from baseplate.lib.config import Endpoint
from baseplate.lib.config import EndpointConfiguration
from baseplate.lib.config import InternetAddress
from baseplate.lib.thrift_pool import ThriftConnectionPool
from baseplate.thrift import BaseplateServiceV2
from baseplate.thrift.ttypes import IsHealthyProbe
from baseplate.thrift.ttypes import IsHealthyRequest


TIMEOUT = 30  # seconds


def check_thrift_service(endpoint: EndpointConfiguration, probe: int) -> None:
    pool = ThriftConnectionPool(endpoint, size=1, timeout=TIMEOUT)
    with pool.connection() as protocol:
        client = BaseplateServiceV2.Client(protocol)
        assert client.is_healthy(
            request=IsHealthyRequest(probe=probe),
        ), f"service indicated unhealthiness in probe {probe}"


def check_http_service(endpoint: EndpointConfiguration, probe: int) -> None:
    if endpoint.family == socket.AF_INET:
        address: InternetAddress = typing.cast(InternetAddress, endpoint.address)
        url = f"http://{address.host}:{address.port}/health?type={probe}"
    elif endpoint.family == socket.AF_UNIX:
        quoted_path = urllib.parse.quote(typing.cast(str, endpoint.address), safe="")
        url = f"http+unix://{quoted_path}/health?type={probe}"
    else:
        raise ValueError(f"unrecognized socket family {endpoint.family!r}")

    session = requests.Session()
    add_unix_socket_support(session)
    response = session.get(url, timeout=TIMEOUT)
    response.raise_for_status()
    response.json()


CHECKERS = {"thrift": check_thrift_service, "wsgi": check_http_service}


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=sys.modules[__name__].__doc__)

    parser.add_argument(
        "type",
        choices=CHECKERS.keys(),
        default="thrift",
        help="The protocol of the service to check.",
    )
    parser.add_argument(
        "endpoint",
        type=Endpoint,
        default=Endpoint("localhost:9090"),
        help="The endpoint to find the service on.",
    )
    parser.add_argument(
        "--probe",
        choices=[probe.lower() for probe in IsHealthyProbe._NAMES_TO_VALUES],
        default="readiness",
        help="The probe to check.",
    )

    return parser.parse_args()


def run_healthchecks() -> None:
    args = parse_args()

    checker = CHECKERS[args.type]
    checker(args.endpoint, IsHealthyProbe._NAMES_TO_VALUES[args.probe.upper()])
    print("OK!")


if __name__ == "__main__":
    run_healthchecks()



import sys
import time

from logging import Logger
from typing import Any
from typing import Callable
from typing import Mapping
from typing import Optional

from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram
from requests.structures import CaseInsensitiveDict
from thrift.protocol.TProtocol import TProtocolBase
from thrift.protocol.TProtocol import TProtocolException
from thrift.Thrift import TApplicationException
from thrift.Thrift import TException
from thrift.Thrift import TProcessor
from thrift.transport.TTransport import TTransportException

from baseplate import Baseplate
from baseplate import RequestContext
from baseplate import TraceInfo
from baseplate.lib.edgecontext import EdgeContextFactory
from baseplate.lib.prometheus_metrics import default_latency_buckets
from baseplate.thrift.ttypes import Error
from baseplate.thrift.ttypes import ErrorCode


PROM_NAMESPACE = "thrift_server"

PROM_LATENCY = Histogram(
    f"{PROM_NAMESPACE}_latency_seconds",
    "Time spent processing requests",
    [
        "thrift_method",
        "thrift_success",
    ],
    buckets=default_latency_buckets,
)
PROM_REQUESTS = Counter(
    f"{PROM_NAMESPACE}_requests_total",
    "Total RPC request count",
    [
        "thrift_method",
        "thrift_success",
        "thrift_exception_type",
        "thrift_baseplate_status",
        "thrift_baseplate_status_code",
    ],
)
PROM_ACTIVE = Gauge(
    f"{PROM_NAMESPACE}_active_requests",
    "The number of in-flight requests being handled by the service",
    ["thrift_method"],
)


class _ContextAwareHandler:
    def __init__(
        self,
        handler: Any,
        context: RequestContext,
        logger: Logger,
        convert_to_baseplate_error: bool,
    ):
        self.handler = handler
        self.context = context
        self.logger = logger
        self.convert_to_baseplate_error = convert_to_baseplate_error

    def __getattr__(self, fn_name: str) -> Callable[..., Any]:
        def call_with_context(*args: Any, **kwargs: Any) -> Any:
            self.logger.debug("Handling: %r", fn_name)

            handler_fn = getattr(self.handler, fn_name)

            span = self.context.span
            span.set_tag("thrift.method", fn_name)
            start_time = time.perf_counter()

            try:
                span.start()
                with PROM_ACTIVE.labels(fn_name).track_inprogress():
                    result = handler_fn(self.context, *args, **kwargs)
            except (TApplicationException, TProtocolException, TTransportException):
                # these are subclasses of TException but aren't ones that
                # should be expected in the protocol
                span.finish(exc_info=sys.exc_info())
                raise
            except Error as exc:
                c = ErrorCode()
                status = c._VALUES_TO_NAMES.get(exc.code, "")
                span.set_tag("exception_type", "Error")
                span.set_tag("thrift.status_code", exc.code)
                span.set_tag("thrift.status", status)
                span.set_tag("success", "false")
                # mark 5xx errors as failures since those are still "unexpected"
                if 500 <= exc.code < 600:
                    span.finish(exc_info=sys.exc_info())
                else:
                    span.finish()
                raise
            except TException as e:
                span.set_tag("exception_type", type(e).__name__)
                span.set_tag("success", "false")
                # this is an expected exception, as defined in the IDL
                span.finish()
                raise
            except Exception:  # noqa: E722
                # the handler crashed (or timed out)!
                span.finish(exc_info=sys.exc_info())
                if self.convert_to_baseplate_error:
                    raise Error(
                        code=ErrorCode.INTERNAL_SERVER_ERROR,
                        message="Internal server error",
                    )
                raise
            else:
                # a normal result
                span.finish()
                return result
            finally:
                thrift_success = "true"
                exception_type = ""
                baseplate_status_code = ""
                baseplate_status = ""
                exc_info = sys.exc_info()
                if exc_info[0] is not None:
                    thrift_success = "false"
                    exception_type = exc_info[0].__name__
                    current_exc = exc_info[1]
                    try:
                        # We want the following code to execute whenever the
                        # service raises an instance of Baseplate's `Error` class.
                        # Unfortunately, we cannot just rely on `isinstance` to do
                        # what we want here because some services compile
                        # Baseplate's thrift file on their own and import `Error`
                        # from that. When this is done, `isinstance` will always
                        # return `False` since it's technically a different class.
                        # To fix this, we optimistically try to access `code` on
                        # `current_exc` and just catch the `AttributeError` if the
                        # `code` attribute is not present.
                        # Note: if the error code was not originally defined in baseplate, or the
                        # name associated with the error was overriden, this cannot reflect that
                        # we will emit the status code in both cases
                        # but the status will be blank in the first case, and the baseplate name
                        # in the second
                        baseplate_status_code = current_exc.code  # type: ignore
                        baseplate_status = ErrorCode()._VALUES_TO_NAMES.get(current_exc.code, "")  # type: ignore
                    except AttributeError:
                        pass
                PROM_REQUESTS.labels(
                    thrift_method=fn_name,
                    thrift_success=thrift_success,
                    thrift_exception_type=exception_type,
                    thrift_baseplate_status=baseplate_status,
                    thrift_baseplate_status_code=baseplate_status_code,
                ).inc()
                PROM_LATENCY.labels(fn_name, thrift_success).observe(
                    time.perf_counter() - start_time
                )

        return call_with_context


def baseplateify_processor(
    processor: TProcessor,
    logger: Logger,
    baseplate: Baseplate,
    edge_context_factory: Optional[EdgeContextFactory] = None,
    convert_to_baseplate_error: bool = False,
) -> TProcessor:
    """Wrap a Thrift Processor with Baseplate's span lifecycle.

    :param processor: The service's processor to wrap.
    :param logger: The logger to use for error and debug logging.
    :param baseplate: The baseplate instance for your application.
    :param edge_context_factory: A configured factory for handling edge request
        context.
    :param convert_to_baseplate_error: If True, the server will automatically
        convert unhandled exceptions to:
            baseplate.Error(
                code=ErrorCode.INTERNAL_SERVER_ERROR,
                message="Internal server error",
            )

    """

    def make_processor_fn(fn_name: str, processor_fn: Callable[..., Any]) -> Callable[..., Any]:
        def call_processor_with_span_context(
            self: Any, seqid: int, iprot: TProtocolBase, oprot: TProtocolBase
        ) -> Any:
            context = baseplate.make_context_object()

            # Allow case-insensitivity for THeader headers
            headers: Mapping[bytes, bytes] = CaseInsensitiveDict(  # type: ignore
                data=iprot.get_headers()
            )

            trace_info: Optional[TraceInfo]
            try:
                sampled = bool(headers.get(b"Sampled") == b"1")
                flags = headers.get(b"Flags", None)
                trace_info = TraceInfo.from_upstream(
                    headers[b"Trace"].decode(),
                    headers[b"Parent"].decode(),
                    headers[b"Span"].decode(),
                    sampled,
                    int(flags) if flags is not None else None,
                )
            except (KeyError, ValueError):
                trace_info = None

            edge_payload = headers.get(b"Edge-Request", None)
            context.raw_edge_context = edge_payload
            if edge_context_factory:
                context.edge_context = edge_context_factory.from_upstream(edge_payload)

            try:
                raw_deadline_budget = headers[b"Deadline-Budget"].decode()
                context.deadline_budget = float(raw_deadline_budget) / 1000
            except (KeyError, ValueError):
                context.deadline_budget = None

            span = baseplate.make_server_span(context, name=fn_name, trace_info=trace_info)
            span.set_tag("protocol", "thrift")
            try:
                service_name = headers[b"User-Agent"].decode()
            except (KeyError, UnicodeDecodeError):
                pass
            else:
                span.set_tag("peer.service", service_name)

            context.headers = headers

            handler = processor._handler
            context_aware_handler = _ContextAwareHandler(
                handler, context, logger, convert_to_baseplate_error
            )
            context_aware_processor = processor.__class__(context_aware_handler)
            return processor_fn(context_aware_processor, seqid, iprot, oprot)

        return call_processor_with_span_context

    instrumented_process_map = {}
    for fn_name, processor_fn in processor._processMap.items():
        context_aware_processor_fn = make_processor_fn(fn_name, processor_fn)
        instrumented_process_map[fn_name] = context_aware_processor_fn
    processor._processMap = instrumented_process_map
    processor.baseplate = baseplate
    return processor



import unittest

try:
    from sqlalchemy import Column, Integer, String
    from sqlalchemy.exc import OperationalError
    from sqlalchemy.ext.declarative import declarative_base
except ImportError:
    raise unittest.SkipTest("sqlalchemy is not installed")

from baseplate.clients.sqlalchemy import (
    engine_from_config,
    SQLAlchemyEngineContextFactory,
    SQLAlchemySession,
    SQLAlchemySessionContextFactory,
)
from baseplate import Baseplate

from . import TestBaseplateObserver


Base = declarative_base()


class TestObject(Base):
    __tablename__ = "test"

    id = Column(Integer, primary_key=True)
    name = Column(String)


class SQLAlchemyEngineTests(unittest.TestCase):
    def setUp(self):
        engine = engine_from_config({"database.url": "sqlite://"})  # in-memory db
        Base.metadata.create_all(bind=engine)
        factory = SQLAlchemyEngineContextFactory(engine)

        self.baseplate_observer = TestBaseplateObserver()

        baseplate = Baseplate()
        baseplate.register(self.baseplate_observer)
        baseplate.add_to_context("db", factory)

        self.context = baseplate.make_context_object()
        self.server_span = baseplate.make_server_span(self.context, "test")

    def test_simple_query(self):
        with self.server_span:
            self.server_span.context.db.execute("SELECT * FROM test;")

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNone(span_observer.on_finish_exc_info)
        span_observer.assert_tag("statement", "SELECT * FROM test;")

    def test_very_long_query(self):
        with self.server_span:
            self.server_span.context.db.execute("SELECT *" + (" " * 1024) + "FROM test;")

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNone(span_observer.on_finish_exc_info)
        tag = span_observer.tags["statement"]
        assert len(tag) <= 1024

    def test_nested_local_span(self):
        with self.server_span:
            with self.server_span.make_child(
                "local", local=True, component_name="example"
            ) as local:
                local.context.db.execute("SELECT * FROM test;")

        server_span_observer = self.baseplate_observer.get_only_child()
        local_span_observer = server_span_observer.get_only_child()
        span_observer = local_span_observer.get_only_child()
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNone(span_observer.on_finish_exc_info)
        span_observer.assert_tag("statement", "SELECT * FROM test;")

    def test_error_in_query(self):
        with self.server_span:
            with self.assertRaises(OperationalError):
                self.context.db.execute("SELECT * FROM does_not_exist;")

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNotNone(span_observer.on_finish_exc_info)


class SQLAlchemySessionTests(unittest.TestCase):
    def setUp(self):
        engine = engine_from_config({"database.url": "sqlite://"})  # in-memory db
        Base.metadata.create_all(bind=engine)
        factory = SQLAlchemySessionContextFactory(engine)

        self.baseplate_observer = TestBaseplateObserver()

        baseplate = Baseplate()
        baseplate.register(self.baseplate_observer)
        baseplate.add_to_context("db", factory)

        self.context = baseplate.make_context_object()
        self.server_span = baseplate.make_server_span(self.context, "test")

    def test_simple_session(self):
        with self.server_span:
            new_object = TestObject(name="cool")
            self.context.db.add(new_object)
            self.context.db.commit()

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNone(span_observer.on_finish_exc_info)


class SQLAlchemySessionConfigTests(unittest.TestCase):
    def test_simple_config(self):
        baseplate = Baseplate({"db.url": "sqlite://"})
        baseplate.configure_context({"db": SQLAlchemySession()})

        context = baseplate.make_context_object()
        with baseplate.make_server_span(context, "test"):
            context.db.execute("SELECT 1;")



from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker


class NoRelativeImportsChecker(BaseChecker):
    __implements__ = IAstroidChecker

    name = "no-relative-imports"
    priority = -1
    msgs = {
        "W5403": (
            "Explicit relative import %r should be %r",
            "explicit-relative-import",
            "All imports should be absolute",
        )
    }

    def __init__(self, linter=None):
        super().__init__(linter)
        self.module = None

    def visit_module(self, node):
        self.module = node

    def leave_module(self, node):
        self.module = None

    def visit_importfrom(self, node):
        if node.level is not None:
            self.add_message(
                "explicit-relative-import",
                node=node,
                args=(
                    "." * node.level + node.modname,
                    self.module.relative_to_absolute_name(node.modname, node.level),
                ),
            )


def register(linter):
    checker = NoRelativeImportsChecker(linter)
    linter.register_checker(checker)



from math import ceil
from time import perf_counter
from typing import Any
from typing import Dict
from typing import Optional

import redis

# redis.client.StrictPipeline was renamed to redis.client.Pipeline in version 3.0
try:
    from redis.client import StrictPipeline as Pipeline  # type: ignore
except ImportError:
    from redis.client import Pipeline

from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram

from baseplate import Span
from baseplate.clients import ContextFactory
from baseplate.lib import config
from baseplate.lib import message_queue
from baseplate.lib import metrics

from baseplate.lib.prometheus_metrics import default_latency_buckets

PROM_PREFIX = "redis_client"
PROM_LABELS_PREFIX = "redis"

PROM_SHARED_LABELS = [
    f"{PROM_LABELS_PREFIX}_command",
    f"{PROM_LABELS_PREFIX}_database",
    f"{PROM_LABELS_PREFIX}_client_name",
    f"{PROM_LABELS_PREFIX}_type",
]
LATENCY_SECONDS = Histogram(
    f"{PROM_PREFIX}_latency_seconds",
    "Latency histogram for calls made by clients",
    [*PROM_SHARED_LABELS, f"{PROM_LABELS_PREFIX}_success"],
    buckets=default_latency_buckets,
)

REQUESTS_TOTAL = Counter(
    f"{PROM_PREFIX}_requests_total",
    "Total number of requests made by client",
    [*PROM_SHARED_LABELS, f"{PROM_LABELS_PREFIX}_success"],
)

ACTIVE_REQUESTS = Gauge(
    f"{PROM_PREFIX}_active_requests",
    "Number of active requests for a given client",
    PROM_SHARED_LABELS,
)

PROM_POOL_PREFIX = f"{PROM_PREFIX}_pool"
PROM_LABELS = ["redis_pool"]

MAX_CONNECTIONS = Gauge(
    f"{PROM_POOL_PREFIX}_max_size",
    "Maximum number of connections allowed in this redis client connection pool",
    PROM_LABELS,
)
IDLE_CONNECTIONS = Gauge(
    f"{PROM_POOL_PREFIX}_idle_connections",
    "Number of idle connections in this redis client connection pool",
    PROM_LABELS,
)
OPEN_CONNECTIONS = Gauge(
    f"{PROM_POOL_PREFIX}_active_connections",
    "Number of open connections in this redis client connection pool",
    PROM_LABELS,
)


def pool_from_config(
    app_config: config.RawConfig, prefix: str = "redis.", **kwargs: Any
) -> redis.ConnectionPool:
    """Make a ConnectionPool from a configuration dictionary.

    The keys useful to :py:func:`pool_from_config` should be prefixed, e.g.
    ``redis.url``, ``redis.max_connections``, etc. The ``prefix`` argument
    specifies the prefix used to filter keys.  Each key is mapped to a
    corresponding keyword argument on the :py:class:`redis.ConnectionPool`
    constructor.

    Supported keys:

    * ``url`` (required): a URL like ``redis://localhost/0``.
    * ``max_connections``: an integer maximum number of connections in the pool
    * ``socket_connect_timeout``: how long to wait for sockets to connect. e.g.
        ``200 milliseconds`` (:py:func:`~baseplate.lib.config.Timespan`)
    * ``socket_timeout``: how long to wait for socket operations, e.g.
        ``200 milliseconds`` (:py:func:`~baseplate.lib.config.Timespan`)

    """
    assert prefix.endswith(".")
    parser = config.SpecParser(
        {
            "url": config.String,
            "max_connections": config.Optional(config.Integer, default=None),
            "socket_connect_timeout": config.Optional(config.Timespan, default=None),
            "socket_timeout": config.Optional(config.Timespan, default=None),
        }
    )
    options = parser.parse(prefix[:-1], app_config)

    if options.max_connections is not None:
        kwargs.setdefault("max_connections", options.max_connections)
    if options.socket_connect_timeout is not None:
        kwargs.setdefault("socket_connect_timeout", options.socket_connect_timeout.total_seconds())
    if options.socket_timeout is not None:
        kwargs.setdefault("socket_timeout", options.socket_timeout.total_seconds())

    return redis.BlockingConnectionPool.from_url(options.url, **kwargs)


class RedisClient(config.Parser):
    """Configure a Redis client.

    This is meant to be used with
    :py:meth:`baseplate.Baseplate.configure_context`.

    :param redis_client_name: The name of this Redis client. Prefer to use `client_name`
    keyword argument to the `pool_from_config` function. Use this if your using a Redis
    host or proxy that doesn't support the `CLIENT SETNAME` function.

    See :py:func:`pool_from_config` for available configuration settings.

    """

    def __init__(self, redis_client_name: str = "", **kwargs: Any):
        # This is for backwards compatibility. Originally we asked clients to
        # set the `client_name` attribute to get the `redis_client_name`
        # tag to appear on Prometheus metrics. Unfortunately this broke clients
        # that use a proxy to connect to Redis.
        # See: https://github.com/redis/redis-py/issues/2384
        client_name = redis_client_name
        if client_name == "":
            client_name = kwargs.get("client_name", "")

        self.kwargs = kwargs
        self.redis_client_name = client_name

    def parse(self, key_path: str, raw_config: config.RawConfig) -> "RedisContextFactory":
        connection_pool = pool_from_config(raw_config, f"{key_path}.", **self.kwargs)
        return RedisContextFactory(
            connection_pool=connection_pool,
            name=key_path,
            redis_client_name=self.redis_client_name,
        )


class RedisContextFactory(ContextFactory):
    """Redis client context factory.

    This factory will attach a
    :py:class:`~baseplate.clients.redis.MonitoredRedisConnection` to an
    attribute on the :py:class:`~baseplate.RequestContext`. When Redis commands
    are executed via this connection object, they will use connections from the
    provided :py:class:`redis.ConnectionPool` and automatically record
    diagnostic information.

    :param connection_pool: A connection pool.

    """

    def __init__(
        self,
        connection_pool: redis.ConnectionPool,
        name: str = "redis",
        redis_client_name: str = "",
    ):
        self.connection_pool = connection_pool
        self.name = name
        self.redis_client_name = redis_client_name

    def report_runtime_metrics(self, batch: metrics.Client) -> None:
        if not isinstance(self.connection_pool, redis.BlockingConnectionPool):
            return

        size = self.connection_pool.max_connections
        open_connections_num = len(self.connection_pool._connections)  # type: ignore
        available = self.connection_pool.pool.qsize()
        in_use = size - available

        MAX_CONNECTIONS.labels(self.name).set(size)
        IDLE_CONNECTIONS.labels(self.name).set(available)
        OPEN_CONNECTIONS.labels(self.name).set(open_connections_num)

        batch.gauge("pool.size").replace(size)
        batch.gauge("pool.in_use").replace(in_use)
        batch.gauge("pool.open_and_available").replace(open_connections_num - in_use)

    def make_object_for_context(self, name: str, span: Span) -> "MonitoredRedisConnection":
        return MonitoredRedisConnection(
            context_name=name,
            server_span=span,
            connection_pool=self.connection_pool,
            redis_client_name=self.redis_client_name,
        )


# pylint: disable=too-many-public-methods
class MonitoredRedisConnection(redis.StrictRedis):
    """Redis connection that collects diagnostic information.

    This connection acts like :py:class:`redis.StrictRedis` except that all
    operations are automatically wrapped with diagnostic collection.

    The interface is the same as that class except for the
    :py:meth:`~baseplate.clients.redis.MonitoredRedisConnection.pipeline`
    method.

    """

    def __init__(
        self,
        context_name: str,
        server_span: Span,
        connection_pool: redis.ConnectionPool,
        redis_client_name: str = "",
    ):
        self.context_name = context_name
        self.server_span = server_span
        self.redis_client_name = redis_client_name

        super().__init__(connection_pool=connection_pool)

    def execute_command(self, *args: Any, **kwargs: Any) -> Any:
        command = args[0]
        trace_name = f"{self.context_name}.{command}"

        labels = {
            f"{PROM_LABELS_PREFIX}_command": command,
            f"{PROM_LABELS_PREFIX}_client_name": self.redis_client_name,
            f"{PROM_LABELS_PREFIX}_database": self.connection_pool.connection_kwargs.get("db", ""),
            f"{PROM_LABELS_PREFIX}_type": "standalone",
        }
        with self.server_span.make_child(trace_name), ACTIVE_REQUESTS.labels(
            **labels
        ).track_inprogress():
            start_time = perf_counter()
            success = "true"

            try:
                res = super().execute_command(command, *args[1:], **kwargs)
                if isinstance(res, redis.RedisError):
                    success = "false"
                return res
            except:  # noqa: E722
                success = "false"
                raise
            finally:
                result_labels = {**labels, f"{PROM_LABELS_PREFIX}_success": success}
                REQUESTS_TOTAL.labels(**result_labels).inc()
                LATENCY_SECONDS.labels(**result_labels).observe(perf_counter() - start_time)

    # pylint: disable=arguments-renamed
    def pipeline(  # type: ignore
        self, name: str, transaction: bool = True, shard_hint: Optional[str] = None
    ) -> "MonitoredRedisPipeline":
        """Create a pipeline.

        This returns an object on which you can call the standard Redis
        commands. Execution will be deferred until ``execute`` is called. This
        is useful for saving round trips.

        :param name: The name to attach to diagnostics for this pipeline.
        :param transaction: Whether or not the commands in the pipeline
            are wrapped with a transaction and executed atomically.

        """
        return MonitoredRedisPipeline(
            f"{self.context_name}.pipeline_{name}",
            self.server_span,
            self.connection_pool,
            self.response_callbacks,
            transaction=transaction,
            shard_hint=shard_hint,
            redis_client_name=self.redis_client_name,
        )

    # these commands are not yet implemented, but probably not unimplementable
    def transaction(self, *args: Any, **kwargs: Any) -> Any:
        """Not currently implemented."""
        raise NotImplementedError


class MonitoredRedisPipeline(Pipeline):
    def __init__(
        self,
        trace_name: str,
        server_span: Span,
        connection_pool: redis.ConnectionPool,
        response_callbacks: Dict,
        redis_client_name: str = "",
        **kwargs: Any,
    ):
        self.trace_name = trace_name
        self.server_span = server_span
        self.redis_client_name = redis_client_name
        super().__init__(connection_pool, response_callbacks, **kwargs)

    # pylint: disable=arguments-differ
    def execute(self, **kwargs: Any) -> Any:
        with self.server_span.make_child(self.trace_name):
            success = "true"
            start_time = perf_counter()
            labels = {
                f"{PROM_LABELS_PREFIX}_command": "pipeline",
                f"{PROM_LABELS_PREFIX}_client_name": self.redis_client_name,
                f"{PROM_LABELS_PREFIX}_database": self.connection_pool.connection_kwargs.get(
                    "db", ""
                ),
                f"{PROM_LABELS_PREFIX}_type": "standalone",
            }

            ACTIVE_REQUESTS.labels(**labels).inc()

            try:
                return super().execute(**kwargs)
            except:  # noqa: E722
                success = "false"
                raise
            finally:
                ACTIVE_REQUESTS.labels(**labels).dec()
                result_labels = {
                    **labels,
                    f"{PROM_LABELS_PREFIX}_success": success,
                }
                REQUESTS_TOTAL.labels(**result_labels).inc()
                LATENCY_SECONDS.labels(**result_labels).observe(perf_counter() - start_time)


class MessageQueue:
    """A Redis-backed variant of :py:class:`~baseplate.lib.message_queue.MessageQueue`.

    :param name: can be any string.

    :param client: should be a :py:class:`redis.ConnectionPool` or
           :py:class:`redis.BlockingConnectionPool` from which a client
           connection can be created from (preferably generated from the
           :py:func:`pool_from_config` helper).

    """

    def __init__(self, name: str, client: redis.ConnectionPool):
        self.queue = name
        if isinstance(client, (redis.BlockingConnectionPool, redis.ConnectionPool)):
            self.client = redis.Redis(connection_pool=client)
        else:
            self.client = client

    def get(self, timeout: Optional[float] = None) -> bytes:
        """Read a message from the queue.

        :param timeout: If the queue is empty, the call will block up to
            ``timeout`` seconds or forever if ``None``, if a float is given,
            it will be rounded up to be an integer
        :raises: :py:exc:`~baseplate.lib.message_queue.TimedOutError` The queue
            was empty for the allowed duration of the call.

        """
        if isinstance(timeout, float):
            timeout = int(ceil(timeout))

        if timeout == 0:
            message = self.client.lpop(self.queue)
        else:
            message = self.client.blpop(self.queue, timeout=timeout or 0)

            if message:
                message = message[1]

        if not message:
            raise message_queue.TimedOutError

        return message

    def put(  # pylint: disable=unused-argument
        self, message: bytes, timeout: Optional[float] = None
    ) -> None:
        """Add a message to the queue.

        :param message: will be typecast to a string upon storage and will come
               out of the queue as a string regardless of what type they are
               when passed into this method.
        """
        self.client.rpush(self.queue, message)

    def unlink(self) -> None:
        """Not implemented for Redis variant."""

    def close(self) -> None:
        """Close queue when finished.

        Will delete the queue from the Redis server (Note, can still enqueue
        and dequeue as the actions will recreate the queue)
        """
        self.client.delete(self.queue)



import base64
import unittest

from unittest import mock

from baseplate.lib.crypto import validate_signature
from baseplate.testing.lib.secrets import FakeSecretsStore


has_csrf_policy = True
try:
    from baseplate.frameworks.pyramid.csrf import _make_csrf_token_payload, TokenCSRFStoragePolicy
except ImportError:
    has_csrf_policy = False


@unittest.skipIf(not has_csrf_policy, "Does not have the required pyramid version")
@mock.patch("baseplate.lib.crypto.time")
class TokenCSRFStoragePolicyTests(unittest.TestCase):
    def setUp(self):
        secrets = FakeSecretsStore(
            {
                "secrets": {
                    "secret/csrf/signing-key": {
                        "type": "versioned",
                        "current": base64.b64encode(b"test"),
                        "encoding": "base64",
                    }
                }
            }
        )
        self.policy = TokenCSRFStoragePolicy(secrets=secrets, secret_path="secret/csrf/signing-key")

    def test_make_csrf_token_payload(self, _):
        prefix, payload = _make_csrf_token_payload(version=1, account_id="t2_1")
        self.assertEqual(prefix, "1")
        self.assertEqual(payload, "1.t2_1")

    def test_new_csrf_token(self, time_mock):
        time_mock.time.return_value = 1000.0
        request = mock.Mock()
        request.authenticated_userid = "t2_1"
        token = self.policy.new_csrf_token(request)
        self.assertTrue(token.startswith("1."))
        self.assertEqual(token, "1.AQAA-BEAAF-br-ovnk0q8Wd0kA98-jsak9elbMqo0WbjT0GuyRTD")
        signature = token.split(".")[-1]
        validate_signature(self.policy._get_secret(), "1.t2_1", signature)

    def test_get_csrf_token(self, _):
        request = mock.Mock()
        request.params = {"csrf_token": "token"}
        self.assertEqual(self.policy.get_csrf_token(request), "token")

    def test_get_csrf_token_missing(self, _):
        request = mock.Mock()
        request.params = {}
        self.assertIs(self.policy.get_csrf_token(request), None)

    def test_check_csrf_token_pass(self, time_mock):
        time_mock.time.return_value = 1000.0
        token = "1.AQAA-BEAAF-br-ovnk0q8Wd0kA98-jsak9elbMqo0WbjT0GuyRTD"
        request = mock.Mock()
        request.authenticated_userid = "t2_1"
        self.assertTrue(self.policy.check_csrf_token(request, token))

    def test_check_csrf_token_tampered(self, time_mock):
        time_mock.time.return_value = 1000.0
        token = "2.AQAA-BEAAF-br-ovnk0q8Wd0kA98-jsak9elbMqo0WbjT0GuyRTD"
        request = mock.Mock()
        request.authenticated_userid = "t2_1"
        self.assertFalse(self.policy.check_csrf_token(request, token))

    def test_check_csrf_token_user_mismatch(self, time_mock):
        time_mock.time.return_value = 1000.0
        token = "1.AQAA-BEAAF-br-ovnk0q8Wd0kA98-jsak9elbMqo0WbjT0GuyRTD"
        request = mock.Mock()
        request.authenticated_userid = "t2_2"
        self.assertFalse(self.policy.check_csrf_token(request, token))

    def test_check_csrf_token_expired(self, time_mock):
        # This signature was generated with a timestamp of 1.0 and an
        # expiration of 1 second, so it expired after
        # datetime.datetime(1970, 1, 1, 0, 0, 2)
        time_mock.time.return_value = 1000.0
        token = "1.AQAAAgAAAEW5P28E0nhnhNrFyQshn9OTlRhOgg3EkjXcNtUXHD9P"
        request = mock.Mock()
        request.authenticated_userid = "t2_2"
        self.assertFalse(self.policy.check_csrf_token(request, token))

    def test_check_csrf_token_null(self, _):
        request = mock.Mock()
        request.authenticated_userid = "t2_1"
        self.assertFalse(self.policy.check_csrf_token(request, None))

    def test_check_csrf_token_invalid(self, time_mock):
        time_mock.time.return_value = 1000.0
        token = "1.foo.AQAA-BEAAF-br-ovnk0q8Wd0kA98-jsak9elbMqo0WbjT0GuyRTD"
        request = mock.Mock()
        request.authenticated_userid = "t2_1"
        self.assertFalse(self.policy.check_csrf_token(request, token))



import builtins
import json
import os
import tempfile
import unittest

from unittest import mock

from baseplate.lib import file_watcher
from baseplate.lib.retry import RetryPolicy


class FileWatcherTests(unittest.TestCase):
    def test_file_not_found_throws_error(self):
        mock_parser = mock.Mock()
        watcher = file_watcher.FileWatcher("/does_not_exist", mock_parser)
        with self.assertRaises(file_watcher.WatchedFileNotAvailableError):
            watcher.get_data()
        self.assertEqual(mock_parser.call_count, 0)

    def test_file_loads_and_parses(self):
        with tempfile.NamedTemporaryFile() as watched_file:
            watched_file.write(b"hello!")
            watched_file.flush()
            mock_parser = mock.Mock()
            watcher = file_watcher.FileWatcher(watched_file.name, mock_parser)

            result = watcher.get_data()
            self.assertEqual(result, mock_parser.return_value)
            self.assertEqual(mock_parser.call_count, 1)

            result = watcher.get_data()
            self.assertEqual(mock_parser.call_count, 1)

        # ensure the loaded data stays around even when the file was deleted
        result = watcher.get_data()
        self.assertEqual(result, mock_parser.return_value)
        self.assertEqual(mock_parser.call_count, 1)

    def test_file_reloads_when_changed(self):
        with tempfile.NamedTemporaryFile() as watched_file:
            watched_file.write(b"hello!")
            watched_file.flush()
            os.utime(watched_file.name, (1, 1))
            watcher = file_watcher.FileWatcher(watched_file.name, parser=lambda x: x.read())

            result = watcher.get_data()
            self.assertEqual(result, "hello!")

            watched_file.seek(0)
            watched_file.write(b"breaking news: hello again!")
            watched_file.flush()
            os.utime(watched_file.name, (2, 2))

            result = watcher.get_data()
            self.assertEqual(result, "breaking news: hello again!")

    def test_file_failing_to_parse_on_first_load_raises(self):
        with tempfile.NamedTemporaryFile() as watched_file:
            watched_file.write(b"!")
            watched_file.flush()
            os.utime(watched_file.name, (1, 1))
            watcher = file_watcher.FileWatcher(watched_file.name, parser=json.load)

            with self.assertRaises(file_watcher.WatchedFileNotAvailableError):
                watcher.get_data()

    def test_file_failing_to_parse_after_first_load_uses_cached_data(self):
        with tempfile.NamedTemporaryFile() as watched_file:
            watched_file.write(b'{"a": 1}')
            watched_file.flush()
            os.utime(watched_file.name, (1, 1))
            watcher = file_watcher.FileWatcher(watched_file.name, parser=json.load)

            result = watcher.get_data()
            self.assertEqual(result, {"a": 1})

            watched_file.seek(0)
            watched_file.write(b"!")
            watched_file.flush()
            os.utime(watched_file.name, (2, 2))

            result = watcher.get_data()
            self.assertEqual(result, {"a": 1})

            watched_file.seek(0)
            watched_file.write(b'{"b": 3}')
            watched_file.flush()
            os.utime(watched_file.name, (3, 3))

            result = watcher.get_data()
            self.assertEqual(result, {"b": 3})

    @mock.patch("baseplate.lib.retry.RetryPolicy.new")
    def test_timeout(self, mock_retry_factory):
        with tempfile.NamedTemporaryFile() as watched_file:
            watched_file.write(b"!")
            watched_file.flush()
            os.utime(watched_file.name, (1, 1))

            mock_retry_policy = mock.MagicMock(spec=RetryPolicy)
            mock_retry_policy.__iter__ = mock.Mock(return_value=iter([3, 2, 1]))
            mock_retry_factory.return_value = mock_retry_policy

            with self.assertRaises(file_watcher.WatchedFileNotAvailableError):
                file_watcher.FileWatcher(watched_file.name, parser=json.load, timeout=3)

    def test_binary_mode(self):
        with tempfile.NamedTemporaryFile() as watched_file:
            watched_file.write(b"foo")
            watched_file.flush()
            os.utime(watched_file.name, (1, 1))

            watcher = file_watcher.FileWatcher(
                watched_file.name, parser=lambda f: f.read(), binary=True
            )

            # mock.mock_open does not appear to work with binary read_data, you
            # end up getting the following error:
            # TypeError: 'str' does not support the buffer interface
            # So all we are really checking is the arguments passed to `open`.
            with mock.patch.object(
                builtins, "open", mock.mock_open(read_data="foo"), create=True
            ) as open_mock:
                watcher.get_data()
            open_mock.assert_called_once_with(
                watched_file.name, encoding=None, mode="rb", newline=None
            )

            watcher = file_watcher.FileWatcher(
                watched_file.name, parser=lambda f: f.read(), binary=True
            )
            self.assertEqual(watcher.get_data(), b"foo")

    def test_text_mode(self):
        with tempfile.NamedTemporaryFile() as watched_file:
            watched_file.write(b"foo")
            watched_file.flush()
            os.utime(watched_file.name, (1, 1))

            watcher = file_watcher.FileWatcher(
                watched_file.name, parser=lambda f: f.read(), binary=False
            )

            with mock.patch.object(
                builtins, "open", mock.mock_open(read_data="foo"), create=True
            ) as open_mock:
                watcher.get_data()
            open_mock.assert_called_once_with(
                watched_file.name, encoding="UTF-8", mode="r", newline=None
            )

            watcher = file_watcher.FileWatcher(
                watched_file.name, parser=lambda f: f.read(), binary=False
            )
            self.assertEqual(watcher.get_data(), "foo")


class Py3FileWatcherTests(unittest.TestCase):
    def test_cant_set_encoding_and_binary(self):
        mock_parser = mock.Mock()
        with self.assertRaises(TypeError):
            file_watcher.FileWatcher("/does_not_exist", mock_parser, binary=True, encoding="utf-8")

    def test_cant_set_newline_and_binary(self):
        mock_parser = mock.Mock()
        with self.assertRaises(TypeError):
            file_watcher.FileWatcher("/does_not_exist", mock_parser, binary=True, newline="\n")

    def test_encoding_option(self):
        file_path = os.path.abspath("tests/data/file_watcher_tests.json")

        watcher = file_watcher.FileWatcher(file_path, parser=json.load, encoding="ANSI_X3.4-1968")
        with self.assertRaises(file_watcher.WatchedFileNotAvailableError):
            watcher.get_data()

        watcher = file_watcher.FileWatcher(file_path, parser=json.load, encoding="utf-8")
        result = watcher.get_data()
        self.assertEqual(result, {"a": "☃️"})

    def test_newline_option(self):
        def parser(f):
            return f.readlines()

        with tempfile.NamedTemporaryFile() as watched_file:
            watched_file.write(b"A\nB\rC\r\nD")
            watched_file.flush()
            os.utime(watched_file.name, (1, 1))

            watcher = file_watcher.FileWatcher(watched_file.name, parser=parser)
            self.assertEqual(watcher.get_data(), ["A\n", "B\n", "C\n", "D"])

            watcher = file_watcher.FileWatcher(watched_file.name, parser=parser, newline="")
            self.assertEqual(watcher.get_data(), ["A\n", "B\r", "C\r\n", "D"])

            watcher = file_watcher.FileWatcher(watched_file.name, parser=parser, newline="\n")
            self.assertEqual(watcher.get_data(), ["A\n", "B\rC\r\n", "D"])

            watcher = file_watcher.FileWatcher(watched_file.name, parser=parser, newline="\r")
            self.assertEqual(watcher.get_data(), ["A\nB\r", "C\r", "\nD"])

            watcher = file_watcher.FileWatcher(watched_file.name, parser=parser, newline="\r\n")
            self.assertEqual(watcher.get_data(), ["A\nB\rC\r\n", "D"])

            watcher = file_watcher.FileWatcher(watched_file.name, parser=parser, newline="foo")
            with self.assertRaises(file_watcher.WatchedFileNotAvailableError):
                watcher.get_data()



import unittest

from contextlib import nullcontext as does_not_raise
from unittest import mock

import pytest

from prometheus_client import REGISTRY
from thrift.protocol.TProtocol import TProtocolException
from thrift.Thrift import TApplicationException
from thrift.Thrift import TException
from thrift.transport.TTransport import TTransportException

from baseplate.clients import thrift
from baseplate.clients.thrift import _build_thrift_proxy_method
from baseplate.clients.thrift import ACTIVE_REQUESTS
from baseplate.clients.thrift import REQUEST_LATENCY
from baseplate.clients.thrift import REQUESTS_TOTAL
from baseplate.clients.thrift import ThriftContextFactory
from baseplate.thrift import BaseplateServiceV2
from baseplate.thrift.ttypes import Error
from baseplate.thrift.ttypes import ErrorCode


class EnumerateServiceMethodsTests(unittest.TestCase):
    def test_enumerate_none(self):
        class Iface:
            pass

        class ExampleClient(Iface):
            pass

        methods = list(thrift._enumerate_service_methods(ExampleClient))

        self.assertEqual(methods, [])

    def test_enumerate_some(self):
        class Iface:
            def some_method(self):
                pass

        class ExampleClient(Iface):
            pass

        methods = list(thrift._enumerate_service_methods(ExampleClient))

        self.assertEqual(set(methods), {"some_method"})

    def test_inherited(self):
        class Iface:
            def local_method(self):
                pass

        class ExampleClient(BaseplateServiceV2.Client, Iface):
            pass

        methods = list(thrift._enumerate_service_methods(ExampleClient))

        self.assertEqual(set(methods), {"is_healthy", "local_method"})

    def test_not_subclass_of_iface(self):
        class ExampleClient:
            pass

        with self.assertRaises(AssertionError):
            list(thrift._enumerate_service_methods(ExampleClient))


class TestPrometheusMetrics:
    def setup(self):
        REQUEST_LATENCY.clear()
        REQUESTS_TOTAL.clear()
        ACTIVE_REQUESTS.clear()

    @pytest.mark.parametrize(
        "exc,exc_type,status,status_code,expectation",
        [
            (None, "", "", "", does_not_raise()),
            (
                TApplicationException(TApplicationException.UNKNOWN_METHOD, "unknown method"),
                "TApplicationException",
                "",
                "",
                pytest.raises(TApplicationException),
            ),
            (
                TProtocolException(message="Required field is unset!"),
                "TProtocolException",
                "",
                "",
                pytest.raises(TProtocolException),
            ),
            (
                TTransportException(message="Something is wrong with the transport"),
                "TTransportException",
                "",
                "",
                pytest.raises(
                    TTransportException, match=r"retry policy exhausted while attempting.*"
                ),
            ),
            (
                Error(ErrorCode.NOT_FOUND, "404 not found"),
                "Error",
                "NOT_FOUND",
                "404",
                pytest.raises(Error),
            ),
            (
                Error(ErrorCode.SERVICE_UNAVAILABLE, "503 unavailable"),
                "Error",
                "SERVICE_UNAVAILABLE",
                "503",
                pytest.raises(Error),
            ),
            (
                TException(message="Some other generic thrift exception"),
                "TException",
                "",
                "",
                pytest.raises(TException),
            ),
            (
                Exception("Some very generic exception"),
                "Exception",
                "",
                "",
                pytest.raises(Exception),
            ),
        ],
    )
    def test_build_thrift_proxy_method(self, exc, exc_type, status, status_code, expectation):
        def handle(*args, **kwargs):
            if exc is None:
                return 42
            else:
                raise exc

        proxy_method = _build_thrift_proxy_method("handle")
        pool = mock.MagicMock(timeout=None)
        prot = mock.MagicMock()
        pool.connection().__enter__.return_value = prot
        client_cls = mock.MagicMock()
        client_cls.handle = handle
        handler = mock.MagicMock(
            retry_policy=[None, None],
            pool=pool,
            namespace="test_namespace",
        )
        handler.client_cls.return_value = client_cls

        thrift_success = str((exc is None)).lower()
        prom_labels = {
            "thrift_method": "handle",
            "thrift_client_name": "test_namespace",
        }
        requests_total_prom_labels = {
            "thrift_exception_type": exc_type,
            "thrift_baseplate_status": status,
            "thrift_baseplate_status_code": status_code,
        }

        with expectation:
            proxy_method(self=handler)

        tries = 1 if exc_type != "TTransportException" else 2
        assert (
            REGISTRY.get_sample_value(
                "thrift_client_requests_total",
                {**prom_labels, **requests_total_prom_labels, "thrift_success": thrift_success},
            )
            == tries
        )
        assert (
            REGISTRY.get_sample_value(
                "thrift_client_latency_seconds_bucket",
                {**prom_labels, "thrift_success": thrift_success, "le": "+Inf"},
            )
            == tries
        )
        assert REGISTRY.get_sample_value("thrift_client_active_requests", prom_labels) == 0

    def test_build_thrift_proxy_method_fail_connection(self):
        def handle(*args, **kwargs):
            return 42

        proxy_method = _build_thrift_proxy_method("handle")
        pool = mock.MagicMock(timeout=None)
        pool.connection().__enter__.side_effect = Exception("failed to establish connection")
        client_cls = mock.MagicMock()
        client_cls.handle = handle
        handler = mock.MagicMock(
            retry_policy=[None, None],
            pool=pool,
            namespace="test_namespace",
        )
        handler.client_cls.return_value = client_cls

        with pytest.raises(Exception):
            proxy_method(self=handler)

        prom_labels = {
            "thrift_method": "handle",
            "thrift_client_name": "test_namespace",
        }
        assert REGISTRY.get_sample_value("thrift_client_active_requests", prom_labels) is None


class TestThriftContextFactory:
    @pytest.fixture
    def pool(self):
        yield mock.MagicMock(size=4, checkedout=8)

    @pytest.fixture
    def context_factory(self, pool):
        class Iface:
            def handle(*args, **kwargs):
                pass

        context_factory = ThriftContextFactory(
            pool=pool,
            client_cls=Iface,
        )
        context_factory.max_connections_gauge.clear()
        context_factory.active_connections_gauge.clear()
        yield context_factory

    def test_thrift_server_pool_prometheus_metrics(self, context_factory):
        context_factory.report_runtime_metrics(batch=mock.MagicMock())
        prom_labels = {
            "thrift_pool": "Iface",
        }
        assert REGISTRY.get_sample_value("thrift_client_pool_max_size", prom_labels) == 4
        assert REGISTRY.get_sample_value("thrift_client_pool_active_connections", prom_labels) == 8



import itertools
import unittest

from unittest import mock

from baseplate.lib.retry import ExponentialBackoffRetryPolicy
from baseplate.lib.retry import IndefiniteRetryPolicy
from baseplate.lib.retry import MaximumAttemptsRetryPolicy
from baseplate.lib.retry import RetryPolicy
from baseplate.lib.retry import TimeBudgetRetryPolicy


class RetryPolicyTests(unittest.TestCase):
    def test_maximum_attempts(self):
        base_policy = mock.MagicMock()
        base_policy.__iter__.return_value = itertools.repeat(1.2)
        policy = MaximumAttemptsRetryPolicy(base_policy, attempts=3)

        retries = iter(policy)
        self.assertEqual(next(retries), 1.2)
        self.assertEqual(next(retries), 1.2)
        self.assertEqual(next(retries), 1.2)
        with self.assertRaises(StopIteration):
            next(retries)

    @mock.patch("time.time", autospec=True)
    def test_time_budget(self, time):
        policy = TimeBudgetRetryPolicy(IndefiniteRetryPolicy(), budget=5)

        time.return_value = 0
        retries = iter(policy)
        self.assertEqual(next(retries), 5)

        time.return_value = 3
        for _ in range(100):
            self.assertEqual(next(retries), 2)

        time.return_value = 7
        with self.assertRaises(StopIteration):
            next(retries)

    @mock.patch("time.time", autospec=True)
    def test_time_budget_always_executes_at_least_once(self, time):
        policy = TimeBudgetRetryPolicy(IndefiniteRetryPolicy(), budget=0)

        time.return_value = 0
        retries = iter(policy)
        self.assertEqual(next(retries), 0)

        with self.assertRaises(StopIteration):
            next(retries)

    @mock.patch("time.sleep", autospec=True)
    def test_exponential_backoff(self, sleep):
        base_policy = mock.MagicMock()
        base_policy.__iter__.return_value = itertools.repeat(0.9)
        policy = ExponentialBackoffRetryPolicy(base_policy, base=0.1)

        retries = iter(policy)
        next(retries)
        self.assertEqual(sleep.call_count, 0)

        next(retries)
        sleep.assert_called_with(0.1)

        next(retries)
        sleep.assert_called_with(0.2)

        next(retries)
        sleep.assert_called_with(0.4)

        next(retries)
        sleep.assert_called_with(0.8)

        # the base policy's lower time remaining takes over here
        next(retries)
        sleep.assert_called_with(0.9)


class ComplexPolicyTests(unittest.TestCase):
    @mock.patch("time.sleep", autospec=True)
    def test_attempts_and_backoff(self, sleep):
        policy = RetryPolicy.new(backoff=0.2, attempts=3)
        retries = iter(policy)

        next(retries)

        next(retries)
        sleep.assert_called_with(0.2)

        next(retries)
        sleep.assert_called_with(0.4)

        with self.assertRaises(StopIteration):
            next(retries)

    @mock.patch("time.time", autospec=True)
    @mock.patch("time.sleep", autospec=True)
    def test_budget_overrides_backoff(self, sleep, time):
        policy = RetryPolicy.new(backoff=0.1, budget=1)

        time.return_value = 0
        retries = iter(policy)
        time_remaining = next(retries)
        self.assertAlmostEqual(time_remaining, 1)
        self.assertEqual(sleep.call_count, 0)

        time.return_value = 0.5
        time_remaining = next(retries)
        self.assertAlmostEqual(time_remaining, 0.4)
        sleep.assert_called_with(0.1)

        time.return_value = 0.9
        time_remaining = next(retries)
        self.assertAlmostEqual(time_remaining, 0)
        self.assertAlmostEqual(sleep.call_args[0][0], 0.1, places=2)

        time.return_value = 1
        with self.assertRaises(StopIteration):
            next(retries)



"""Integration with Synapse's ``file_output`` service discovery method.

.. note:: Production Baseplate services have Synapse hooked up to a
    local HAProxy instance which will automatically route connections to
    services for you if you connect to the correct address/port on
    localhost. That is the preferred method of connecting to services.

    The contents of this module are useful for inspecting the service
    inventory directly for cases where a blind TCP connection is
    insufficient (e.g. to give service addresses to a client, or for
    topology-aware clients like Cassandra).

A basic example of usage::

    inventory = ServiceInventory("/var/lib/synapse/example.json")
    backend = inventory.get_backend()
    print(backend.endpoint.address)

"""
import json

from typing import IO
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Sequence

from baseplate.lib.config import Endpoint
from baseplate.lib.config import EndpointConfiguration
from baseplate.lib.file_watcher import FileWatcher
from baseplate.lib.file_watcher import WatchedFileNotAvailableError
from baseplate.lib.random import WeightedLottery


class Backend(NamedTuple):
    """A description of a service backend.

    This is a tuple of several values:

    ``id``
        A unique integer ID identifying the backend.

    ``name``
        The name of the backend.

    ``endpoint``
        An :py:class:`~baseplate.lib.config.EndpointConfiguration` object
        describing the network address of the backend.

    ``weight``
        An integer weight indicating how much to prefer this backend
        when choosing whom to connect to.

    """

    id: int
    name: str
    endpoint: EndpointConfiguration
    weight: int


class _Inventory(NamedTuple):
    backends: List[Backend]
    lottery: Optional[WeightedLottery[Backend]]


def _parse(watched_file: IO) -> _Inventory:
    backends = []
    for d in json.load(watched_file):
        endpoint = Endpoint(f"{d['host']}:{d['port']:d}")
        weight = d["weight"] if d["weight"] is not None else 1
        backend = Backend(d["id"], d["name"], endpoint, weight)
        backends.append(backend)

    lottery = None
    if backends:
        lottery = WeightedLottery(backends, weight_key=lambda b: b.weight)

    return _Inventory(backends, lottery)


class NoBackendsAvailableError(Exception):
    """Raised when no backends are available for this service."""


class ServiceInventory:
    """The inventory enumerates available backends for a single service.

    :param filename: The absolute path to the Synapse-generated inventory file
        in JSON format.

    """

    def __init__(self, filename: str):
        self._filewatcher = FileWatcher(filename, _parse)

    def get_backends(self) -> Sequence[Backend]:
        """Return a list of all available backends in the inventory.

        If the inventory file becomes unavailable, the previously seen
        inventory is returned.

        """
        try:
            # pylint: disable=maybe-no-member
            return self._filewatcher.get_data().backends
        except WatchedFileNotAvailableError:
            return []

    def get_backend(self) -> Backend:
        """Return a randomly chosen backend from the available backends.

        If weights are specified in the inventory, they will be
        respected when making the random selection.

        :raises: :py:exc:`NoBackendsAvailableError` if the inventory
            has no available endpoints.

        """
        inventory: Optional[_Inventory]

        try:
            inventory = self._filewatcher.get_data()
        except WatchedFileNotAvailableError:
            inventory = None

        # pylint: disable=maybe-no-member
        if not inventory or not inventory.lottery:
            raise NoBackendsAvailableError

        return inventory.lottery.pick()



# Libraries needed for tests
import astroid
import pylint.testutils

from baseplate.lint import example_plugin


# CheckerTestCase creates a linter that will traverse the AST tree
class TestNoReassignmentChecker(pylint.testutils.CheckerTestCase):
    CHECKER_CLASS = example_plugin.NoReassignmentChecker

    # Use astroid.extract_node() to create a test case
    # Where you put #@ is where the variable gets assigned
    # example, assign_node_a = test = 1, assign_node_b = test = 2
    def test_finds_reassigned_variable(self):
        assign_node_a, assign_node_b = astroid.extract_node(
            """
        test = 1 #@
        test = 2 #@
            """
        )

        self.checker.visit_assign(assign_node_a)
        self.checker.visit_assign(assign_node_b)
        self.assertAddsMessages(
            pylint.testutils.MessageTest(msg_id="reassigned-variable", node=assign_node_a)
        )

    def test_ignores_no_reassigned_variable(self):
        assign_node_a, assign_node_b = astroid.extract_node(
            """
        test1 = 1 #@
        test2 = 2 #@
            """
        )

        with self.assertNoMessages():
            self.checker.visit_assign(assign_node_a)
            self.checker.visit_assign(assign_node_b)

    def test_ignores_variable_outside_function(self):
        func_node, assign_node_a, assign_node_b = astroid.extract_node(
            """
        def test1(): #@
            test = 1 #@

        def test2():
            test = 2 #@
            """
        )

        with self.assertNoMessages():
            self.checker.visit_assign(assign_node_a)
            self.checker.leave_functiondef(func_node)
            self.checker.visit_assign(assign_node_b)
